import abc
import logging
from typing import (
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)

from jax import (
    lax,
    random,
    vmap,
)
import jax.numpy as jnp

from fortuna.conformal.multivalid.base import MultivalidMethod
from fortuna.typing import Array


class IterativeMultivalidMethod(MultivalidMethod):
    def __init__(self, seed: int = 0):
        """
        A base iterative multivalid method.

        Parameters
        ----------
        seed: int
            Random seed.
        """
        super().__init__(seed=seed)
        self._patches = []
        self._eta = None
        self._bucket_types = None

    def calibrate(
        self,
        scores: Array,
        groups: Optional[Array] = None,
        values: Optional[Array] = None,
        test_groups: Optional[Array] = None,
        test_values: Optional[Array] = None,
        atol: float = 1e-4,
        rtol: float = 1e-6,
        min_prob_b: Union[float, str] = "auto",
        n_buckets: int = 100,
        n_rounds: int = 1000,
        eta: float = 0.1,
        split: float = 0.8,
        bucket_types: Tuple[str, ...] = ("<=", ">="),
        **kwargs,
    ) -> Union[Dict, Tuple[Array, Dict]]:
        """
        Calibrate the model by finding a list of patches to the model that bring the calibration error below a
        certain threshold.

        Parameters
        ----------
        scores: Array
            A list of scores :math:`s(x, y)` computed on the calibration data.
            This should be a one-dimensional array of elements between 0 and 1.
        groups: Array
            A list of groups :math:`g(x)` computed on the calibration data.
            This should be a two-dimensional array of bool elements.
            The first dimension is over the data points, the second dimension is over the number of groups.
        values: Optional[Array]
            The initial model evaluations :math:`f(x)` on the calibration data. If not provided, these are set to 0.
        test_groups: Optional[Array]
            A list of groups :math:`g(x)` computed on the test data.
            This should be a two-dimensional array of bool elements.
            The first dimension is over the data points, the second dimension is over the number of groups.
        test_values: Optional[Array]
            The initial model evaluations :math:`f(x)` on the test data. If not provided, these are set to 0.
        atol: float
            Absolute tolerance on the loss.
        rtol: float
            Relative tolerance on the loss.
        min_prob_b: Union[float, str]
            Minimum probability of the conditioning set :math:`B_t` for the patch to be applied.
            If "auto", it will be chosen based on the number of buckets and dimension of the scores.
        n_buckets: int
            The number of buckets used in the algorithm.
        n_rounds: int
            The maximum number of rounds to run the method for.
        eta: float
            Step size. By default, this is set to 1.
        split: float
            Split the calibration data into calibration and validation, according to the given proportion.
            The validation data will be used for early stopping.
        bucket_types: Tuple[str, ...]
            Types of buckets. The following types are currently supported:

            - "=", corresponding of buckets like :math:`\{f(x) = v\}`;
            - ">=", corresponding of buckets like :math:`\{f(x) \ge v\}`;
            - "<=", corresponding of buckets like :math:`\{f(x) \le v\}`.

        Returns
        -------
        Union[Dict, Tuple[Array, Dict]]
            A status including the number of rounds taken to reach convergence and the calibration errors computed
            during the training procedure. if `test_values` and `test_groups` are provided, the list of patches will
            be applied to `test_values`, and the calibrated test values will be returned together with the status.
        """
        if n_rounds < 1:
            raise ValueError("`n_rounds` must be at least 1.")

        if test_values is not None and values is None:
            raise ValueError(
                "If `test_values is provided, `values` must also be provided."
            )
        if values is not None and test_groups is not None and test_values is None:
            raise ValueError(
                "If `values` and `test_groups` are provided, `test_values` must also be provided."
            )
        if test_groups is not None and groups is None:
            raise ValueError(
                "If `test_groups` is provided, `groups` must also be provided."
            )
        if groups is not None and test_values is not None and test_groups is None:
            raise ValueError(
                "If `groups` and `test_values` are provided, `test_groups` must also be provided."
            )
        if eta <= 0 or eta > 1:
            raise ValueError(
                "`eta` must be a float greater than 0 and less than or equal to 1."
            )
        if split <= 0 or split > 1:
            raise ValueError("`split` must be a float greater than 0 and less than 1.")
        if min_prob_b != "auto" and (min_prob_b < 0 or min_prob_b > 1):
            raise ValueError(
                "`min_prob_b` must be greater than or equal to 0 and less than or equal to 1."
            )

        self._check_scores(scores)
        scores = self._process_scores(scores)
        n_dims = scores.shape[1]

        min_prob_b = self._maybe_init_min_prob_b(
            min_prob_b=min_prob_b, n_buckets=n_buckets, n_dims=n_dims
        )

        groups = self._init_groups(groups, scores.shape[0])
        self._maybe_check_groups(groups, test_groups)
        n_groups = groups.shape[1]

        values = self._maybe_init_values(values, groups.shape[0])
        self._maybe_check_values(values, test_values)

        self.n_buckets = n_buckets
        buckets = self._get_buckets(n_buckets)
        values = vmap(lambda v: self._round_to_buckets(v, buckets))(values)

        self._check_bucket_types(bucket_types)
        taus = self._get_bucket_type_indices(bucket_types)
        self._bucket_types = bucket_types
        n_bucket_types = len(bucket_types)

        self._eta = eta

        size = len(scores)
        calib_size = int(jnp.ceil(split * size))

        perm = random.choice(
            random.PRNGKey(self._seed), size, shape=(size,), replace=False
        )
        scores, val_scores = scores[perm[:calib_size]], scores[perm[calib_size:]]
        values, val_values = values[perm[:calib_size]], values[perm[calib_size:]]
        groups, val_groups = groups[perm[:calib_size]], groups[perm[calib_size:]]

        val_losses = [float(self._loss_fn(val_values, val_scores))]
        old_losses = float(self._loss_fn(values, scores))
        self._patches = []
        converged = False

        for t in range(n_rounds):
            calib_error_tau_g_v_c, b = vmap(
                lambda tau: vmap(
                    lambda g: vmap(
                        lambda v: vmap(
                            lambda c: self._calibration_error(
                                v=v,
                                g=g,
                                c=c,
                                tau=tau,
                                scores=scores[:, c],
                                groups=groups,
                                values=values,
                                buckets=buckets,
                                **kwargs,
                            )
                        )(jnp.arange(n_dims))
                    )(buckets)
                )(jnp.arange(n_groups))
            )(taus)

            cond_b = jnp.mean(b, axis=-1) >= min_prob_b

            if not jnp.count_nonzero(cond_b):
                logging.warning(
                    f"Early stopping triggered after {t} rounds. "
                    f"No further sets with probability at least `min_prob_b` left available."
                )
                break
            calib_error_tau_g_v_c *= cond_b

            taut, gt, vt, ct, bt = self._get_taut_and_gt_and_vt_and_ct_and_bt(
                calib_error_tau_g_v_c=calib_error_tau_g_v_c,
                buckets=buckets,
                n_groups=n_groups,
                n_dims=n_dims,
                b=b,
                n_bucket_types=n_bucket_types,
                taus=taus,
            )

            patch = self._get_patch(
                v=vt,
                g=gt,
                c=ct,
                b=bt,
                tau=taut,
                scores=scores[:, ct],
                groups=groups,
                values=values,
                buckets=buckets,
                **kwargs,
            )
            values = self._patch(values=values, patch=patch, b=bt, c=ct, eta=eta)

            val_bt = self._get_b(
                groups=val_groups,
                values=val_values,
                v=vt,
                g=gt,
                c=ct,
                tau=taut,
                n_buckets=self.n_buckets,
            )
            val_values = self._patch(
                values=val_values, patch=patch, b=val_bt, c=ct, eta=self.eta
            )

            losses = float(self._loss_fn(values, scores))
            val_losses.append(float(self._loss_fn(val_values, val_scores)))
            if losses > old_losses:
                logging.warning(
                    "The algorithm cannot achieve the desired tolerance. "
                    "Please try increasing `n_buckets`, or decreasing `min_prob_b`."
                )
                break
            if val_losses[-1] <= atol:
                logging.info(f"Absolute tolerance satisfied after {t} rounds.")
                converged = True
                break
            if val_losses[-1] > val_losses[-2]:
                logging.warning(
                    f"Early stopping triggered after {t} rounds. "
                    f"The loss started increasing on the validation data."
                )
                break
            if (
                jnp.any(val_bt)
                and (val_losses[-2] - val_losses[-1]) / val_losses[-2] <= rtol
            ):
                logging.info(f"Relative tolerance satisfied after {t} rounds.")
                converged = True
                break
            old_losses = jnp.copy(losses)

            self._patches.append((taut, gt, vt, ct, patch))

        status = dict(
            n_rounds=len(self.patches),
            losses=val_losses,
            converged=converged,
        )

        if t == n_rounds - 1 and not converged:
            logging.warning(
                "Maximum number of rounds reached without convergence. "
                "Consider adjusting the hyperparameters. "
                "In particular: "
                "increasing `n_rounds`, increasing `n_buckets`, decreasing `min_prob_b` and increasing `eta`."
            )

        if test_groups is not None or test_values is not None:
            test_values = self.apply_patches(test_groups, test_values)
            return test_values, status
        return status

    def apply_patches(
        self,
        groups: Optional[Array] = None,
        values: Optional[Array] = None,
    ) -> Array:
        """
        Apply the patches to the model evaluations.

        Parameters
        ----------
        groups: Array
            A list of groups :math:`g(x)` evaluated over some inputs.
            This should be a two-dimensional array of bool elements.
            The first dimension is over the data points, the second dimension is over the number of groups.
        values: Optional[Array]
            The initial model evaluations :math:`f(x)` evaluated over some inputs. If not provided, these are set to 0.

        Returns
        -------
        Array
            The calibrated values.
        """
        if groups is None and values is None:
            raise ValueError(
                "At least one between `groups` and `values` must be provided."
            )
        if not len(self._patches):
            logging.warning("No patches available.")
            return values

        values = self._maybe_init_values(
            values, groups.shape[0] if groups is not None else None
        )
        self._maybe_check_values(values)

        groups = self._init_groups(groups, values.shape[0])
        self._maybe_check_groups(groups)

        buckets = self._get_buckets(n_buckets=self.n_buckets)
        values = vmap(lambda v: self._round_to_buckets(v, buckets))(values)

        for taut, gt, vt, ct, patch in self._patches:
            bt = self._get_b(
                groups=groups,
                values=values,
                v=vt,
                g=gt,
                c=ct,
                tau=taut,
                n_buckets=self.n_buckets,
            )
            values = self._patch(values=values, patch=patch, b=bt, c=ct, eta=self.eta)
        return values

    def calibration_error(
        self,
        scores: Array,
        groups: Optional[Array] = None,
        values: Optional[Array] = None,
        n_buckets: int = 10000,
        **kwargs,
    ) -> Array:
        """
        The reweighted average squared calibration error :math:`\mu(g) K_2(f, g, \mathcal{D})`.

        Parameters
        ----------
        scores: Array
            A score for each data point.
        groups: Array
            A list of groups :math:`g(x)` evaluated over some inputs.
            This should be a two-dimensional array of bool elements.
            The first dimension is over the data points, the second dimension is over the number of groups.
        values: Array
            The model evaluations, before or after calibration.
        n_buckets: int
            The number of buckets used in the algorithm.

        Returns
        -------
        Array
            The computed calibration error for each group
        """
        self._check_scores(scores)
        scores = self._process_scores(scores)
        n_dims = scores.shape[1]

        self._maybe_check_groups(groups)
        self._maybe_check_values(values)

        groups = self._init_groups(groups, scores.shape[0])
        values = self._maybe_init_values(values, scores.shape[0])

        buckets = self._get_buckets(n_buckets)
        values = vmap(lambda v: self._round_to_buckets(v, buckets))(values)

        error, b = vmap(
            lambda g: vmap(
                lambda v: vmap(
                    lambda c: self._calibration_error(
                        v=v,
                        g=g,
                        c=c,
                        tau=0,
                        scores=scores[:, c],
                        groups=groups,
                        values=values,
                        buckets=buckets,
                        **kwargs,
                    )
                )(jnp.arange(n_dims))
            )(buckets)
        )(jnp.arange(groups.shape[1]))
        error = error.sum(1)

        if n_dims == 1:
            error = error.squeeze(1)
        return error

    @property
    def eta(self):
        return self._eta

    @eta.setter
    def eta(self, eta):
        self._eta = eta

    def _maybe_init_values(self, values: Optional[Array], size: Optional[int] = None):
        if values is None:
            if size is None:
                raise ValueError(
                    "If `values` is not provided, `size` must be provided."
                )
            return jnp.zeros(size)
        return jnp.copy(values)

    @abc.abstractmethod
    def _calibration_error(
        self,
        v: Array,
        g: Array,
        c: Optional[Array],
        tau: Array,
        scores: Array,
        groups: Array,
        values: Array,
        buckets: Array,
        **kwargs,
    ):
        pass

    @staticmethod
    def _get_taut_and_gt_and_vt_and_ct_and_bt(
        calib_error_tau_g_v_c: Array,
        buckets: Array,
        n_groups: int,
        n_dims: int,
        n_bucket_types: int,
        b: Array,
        taus: Array,
    ) -> Tuple[Array, Array, Array, Array, Array]:
        idx_taut, gt, idx_vt, ct = jnp.unravel_index(
            jnp.argmax(calib_error_tau_g_v_c),
            (n_bucket_types, n_groups, len(buckets), n_dims),
        )
        vt = buckets[idx_vt]
        taut = taus[idx_taut]
        return taut, gt, vt, ct, b[idx_taut, gt, idx_vt, ct]

    @staticmethod
    def _get_b(
        groups: Array,
        values: Array,
        v: Array,
        g: Array,
        c: Optional[Array],
        tau: Array,
        n_buckets: int,
    ) -> Array:
        b = lax.select(
            tau == 0,
            jnp.abs(values - v) < 0.5 / n_buckets,
            lax.select(
                tau == 1, values + 0.5 / n_buckets >= v, values - 0.5 / n_buckets <= v
            ),
        )
        b *= groups[:, g]
        return b

    @abc.abstractmethod
    def _get_patch(
        self,
        v: Array,
        g: Array,
        c: Array,
        b: Optional[Array],
        tau: Optional[Array],
        scores: Array,
        groups: Array,
        values: Array,
        buckets: Array,
        **kwargs,
    ) -> Array:
        pass

    def _patch(
        self,
        values: Array,
        patch: Array,
        b: Array,
        c: Array,
        eta: float,
    ) -> Array:
        taken_values_to_set = self._take_values_to_set(values, b, c)
        taken_values = self._take_values(values, b, c)
        return taken_values_to_set.set(jnp.clip(taken_values + eta * patch, 0, 1))

    @staticmethod
    def _take_values(values: Array, b: Array, c: Array):
        if values.ndim == 1:
            return values[b]
        return values[b, c]

    @staticmethod
    def _take_values_to_set(values: Array, b: Array, c: Array):
        if values.ndim == 1:
            return values.at[b]
        return values.at[b, c]

    @staticmethod
    def _loss_fn(values: Array, scores: Array) -> Array:
        pass

    @staticmethod
    def _maybe_check_groups(groups: Array, test_groups: Optional[Array] = None):
        if groups is not None and groups.ndim != 2:
            raise ValueError("`groups` must be a 2-dimensional array.")
        if test_groups is not None and test_groups.ndim != 2:
            raise ValueError("`groups` must be a 2-dimensional array.")
        if groups is not None and groups.dtype != bool:
            raise ValueError("All elements in `groups` must be a bool.")
        if test_groups is not None and test_groups.dtype != bool:
            raise ValueError("All elements in `test_groups` must be a bool.")

    @staticmethod
    def _init_groups(groups: Optional[Array], size: int) -> Array:
        if groups is None:
            return jnp.ones((size, 1), dtype=bool)
        return jnp.copy(groups)

    @property
    def bucket_types(self) -> Tuple[str]:
        return self._bucket_types

    @staticmethod
    def _get_bucket_type_indices(bucket_types: Tuple[str]) -> Array:
        taus = []
        if "=" in bucket_types:
            taus.append(0)
        if ">=" in bucket_types:
            taus.append(1)
        if "<=" in bucket_types:
            taus.append(2)
        return jnp.array(taus)

    @staticmethod
    def _check_bucket_types(bucket_types: Tuple[str]):
        supported_bucket_types = ["=", ">=", "<="]
        for bucket_type in bucket_types:
            if bucket_type not in supported_bucket_types:
                raise ValueError(f"`bucket_type='{bucket_type}'` not recognized.")



from fortuna.prob_model.posterior.base import PosteriorApproximator
from fortuna.prob_model.posterior.normalizing_flow.advi import ADVI_NAME


class ADVIPosteriorApproximator(PosteriorApproximator):
    def __init__(
        self,
        std_init_params: float = 0.1,
        log_std_base: float = -2.3,
        n_loss_samples: int = 3,
    ):
        """
        Automatic Differentiation Variational Inference (ADVI) approximator. It is responsible to define how the
        posterior distribution is approximated.

        Parameters
        ----------
        std_init_params : float
            The standard deviation of the Gaussian distribution used to initialize the parameters of the flow.
        log_std_base : float
            The normalizing flow transforms a base distribution into an approximation of the posterior. The base
            distribution is assumed to be an isotropic Gaussian, with this argument as the log-standard deviation.
        n_loss_samples : int
            Number of samples to approximate the loss, that is the KL divergence (or the ELBO, equivalently).
        """
        self.std_init_params = std_init_params
        self.log_std_base = log_std_base
        self.n_loss_samples = n_loss_samples

    def __str__(self):
        return ADVI_NAME



from fortuna.prob_model.posterior.posterior_mixin import WithPosteriorCheckpointingMixin
from fortuna.training.trainer import TrainerABC


class PosteriorTrainerABC(WithPosteriorCheckpointingMixin, TrainerABC):
    pass



from typing import (
    NamedTuple,
    Optional,
)

import jax
from jax._src.prng import PRNGKeyArray
import jax.numpy as jnp
from optax import GradientTransformation
from optax._src.base import PyTree

from fortuna.prob_model.posterior.sgmcmc.sgmcmc_preconditioner import (
    Preconditioner,
    PreconditionerState,
)
from fortuna.prob_model.posterior.sgmcmc.sgmcmc_step_schedule import StepSchedule
from fortuna.typing import Array
from fortuna.utils.random import generate_random_normal_like_tree


class OptaxSGHMCState(NamedTuple):
    """Optax state for the SGHMC integrator."""

    count: Array
    rng_key: PRNGKeyArray
    momentum: PyTree
    preconditioner_state: PreconditionerState


def sghmc_integrator(
    momentum_decay: float,
    momentum_resample_steps: Optional[int],
    rng_key: PRNGKeyArray,
    step_schedule: StepSchedule,
    preconditioner: Preconditioner,
) -> GradientTransformation:
    """Optax implementation of the SGHMC integrator.

    Parameters
    ----------
        momentum_decay: float
            The momentum decay parameter.
        rng_key: PRNGKeyArray
            An initial random number generator.
        step_schedule: StepSchedule
            A function that takes training step as input and returns the step size.
        preconditioner: Preconditioner
            See :class:`Preconditioner` for reference.
    """
    # Implementation was partially adapted from https://github.com/google-research/google-research/blob/master/bnn_hmc/core/sgmcmc.py#L56

    def init_fn(params):
        return OptaxSGHMCState(
            count=jnp.zeros([], jnp.int32),
            rng_key=rng_key,
            momentum=jax.tree_util.tree_map(jnp.zeros_like, params),
            preconditioner_state=preconditioner.init(params),
        )

    def update_fn(gradient, state, *_):
        step_size = step_schedule(state.count)

        preconditioner_state = preconditioner.update_preconditioner(
            gradient, state.preconditioner_state
        )

        key, new_key = jax.random.split(state.rng_key)
        noise = generate_random_normal_like_tree(key, gradient)
        noise = preconditioner.multiply_by_m_sqrt(noise, preconditioner_state)

        momentum = jax.lax.cond(
            momentum_resample_steps is not None
            and state.count % momentum_resample_steps == 0,
            lambda: jax.tree_util.tree_map(jnp.zeros_like, gradient),
            lambda: state.momentum,
        )

        momentum = jax.tree_map(
            lambda m, g, n: momentum_decay * m
            + g * jnp.sqrt(step_size)
            + n * jnp.sqrt(2 * (1 - momentum_decay)),
            momentum,
            gradient,
            noise,
        )
        updates = preconditioner.multiply_by_m_inv(momentum, preconditioner_state)
        updates = jax.tree_map(lambda m: m * jnp.sqrt(step_size), updates)
        return updates, OptaxSGHMCState(
            count=state.count + 1,
            rng_key=new_key,
            momentum=momentum,
            preconditioner_state=preconditioner_state,
        )

    return GradientTransformation(init_fn, update_fn)



"""
The code has been taken from https://github.com/google/edward2/blob/main/edward2/jax/nn/normalization.py
"""
import dataclasses
from typing import (
    Any,
    Callable,
    Mapping,
    Optional,
    Tuple,
    Type,
)

import flax.core
import flax.linen as nn
import jax
import jax.numpy as jnp
from jax.random import PRNGKeyArray
import numpy as np

from fortuna.typing import (
    Array,
    Shape,
)


def _l2_normalize(x: Array, eps: float = 1e-12) -> Array:
    return x * jax.lax.rsqrt(jnp.maximum(jnp.square(x).sum(), eps))


class SpectralNormalization(nn.Module):
    """
    Implements spectral normalization for linear layers.

    See `Spectral Normalization for Generative Adversarial Networks <https://arxiv.org/abs/1802.05957>`_ .

    Attributes
    ----------
    layer: nn.Module
        A Flax layer to apply normalization to.
    iteration: int
        The number of power iterations to estimate weight matrix's singular value.
    norm_multiplier: float
        Multiplicative constant to threshold the normalization.
        Usually under normalization, the singular value will converge to this value.
    u_init: Callable[[PRNGKeyArray, Shape, Type], Array]
        Initializer function for the first left singular vectors of the kernel.
    v_init: Callable[[PRNGKeyArray, Shape, Type], Array]
        Initializer function for the first right singular vectors of the kernel.
    kernel_apply_kwargs: Optional[Mapping[str, Any]]
        Updated keyword arguments to clone the input layer.
        The cloned layer represents the linear operator performed by the weight matrix.
        If not specified, that operator follows SN-GAN implementation
        (`Takeru M. et al <https://arxiv.org/abs/1802.05957>`_).
        In particular, for Dense layers the default behavior is equivalent to using a cloned layer with no bias
        (by specifying `kernel_apply_kwargs=dict(use_bias=False)`). With this customization, we
        can have the same implementation (inspried by
        (`Stephan H., 2020 <https://nbviewer.jupyter.org/gist/shoyer/fa9a29fd0880e2e033d7696585978bfc>`_)
        for different interpretations of Conv layers. Also see `SpectralNormalizationConv2D` for
        an example of using this attribute.
    kernel_name: str
        Name of the kernel parameter of the input layer.
    layer_name: Optional[str]
        Name of the input layer
    update_singular_value_estimate: Optional[bool]
        Whether to perform power iterations to update the singular value estimate.
    """

    layer: nn.Module
    iteration: int = 1
    norm_multiplier: float = 0.95
    u_init: Callable[[PRNGKeyArray, Shape, Type], Array] = nn.initializers.normal(
        stddev=0.05
    )
    v_init: Callable[[PRNGKeyArray, Shape, Type], Array] = nn.initializers.normal(
        stddev=0.05
    )
    kernel_apply_kwargs: Optional[Mapping[str, Any]] = None
    kernel_name: str = "kernel"
    layer_name: Optional[str] = None
    update_singular_value_estimate: Optional[bool] = None

    def _get_singular_vectors(
        self, initializing: bool, kernel_apply: Callable, in_shape: Shape, dtype: Type
    ) -> Tuple[nn.Variable, nn.Variable]:
        if initializing:
            rng_u = self.make_rng("params")
            rng_v = self.make_rng("params")
            # Interpret output shape (not that this does not cost any FLOPs).
            out_shape = jax.eval_shape(
                kernel_apply, jax.ShapeDtypeStruct(in_shape, dtype)
            ).shape
        else:
            rng_u = rng_v = out_shape = None
        u = self.variable("spectral_stats", "u", self.u_init, rng_u, out_shape, dtype)
        v = self.variable("spectral_stats", "v", self.v_init, rng_v, in_shape, dtype)
        return u, v

    @nn.compact
    def __call__(
        self, inputs: Array, update_singular_value_estimate: Optional[bool] = None
    ) -> Array:
        """
        Applies a linear transformation with spectral normalization to the inputs.

        Parameters
        ----------
        inputs: Array
            The nd-array to be transformed.
        update_singular_value_estimate: Optional[bool]
            Whether to perform power iterations to update the singular value estimate.

        Returns
        -------
        Array
            The transformed input.
        """
        update_singular_value_estimate = nn.merge_param(
            "update_singular_value_estimate",
            self.update_singular_value_estimate,
            update_singular_value_estimate,
        )
        layer_name = (
            type(self.layer).__name__ if self.layer_name is None else self.layer_name
        )
        params = self.param(
            layer_name, lambda *args: self.layer.init(*args)["params"], inputs
        )
        w = params[self.kernel_name]

        if self.kernel_apply_kwargs is None:
            # By default, we use the implementation in SN-GAN.
            kernel_apply = lambda x: x @ w.reshape(-1, w.shape[-1])
            in_shape = (np.prod(w.shape[:-1]),)
        else:
            # Otherwise, we extract the actual kernel transformation in the input
            # layer. This is useful for Conv2D spectral normalization in
            # [Farzan F. et al., 2019](https://arxiv.org/abs/1811.07457).
            kernel_apply = self.layer.clone(
                **self.kernel_apply_kwargs
            ).bind(  # pylint: disable=not-a-mapping
                {"params": {self.kernel_name: w}}
            )
            # Compute input shape of the kernel operator. This is correct for all
            # linear layers on Flax: Dense, Conv, Embed.
            in_shape = inputs.shape[-w.ndim + 1 : -1] + w.shape[-2:-1]

        initializing = self.is_mutable_collection("params")
        u, v = self._get_singular_vectors(initializing, kernel_apply, in_shape, w.dtype)
        u_hat, v_hat = u.value, v.value
        u_, kernel_transpose = jax.vjp(kernel_apply, v_hat)
        if update_singular_value_estimate and not initializing:
            # Run power iterations using autodiff approach inspired by
            # (Stephan H., 2020)[https://nbviewer.jupyter.org/gist/shoyer/fa9a29fd0880e2e033d7696585978bfc]).
            def scan_body(carry, _):
                u_hat, v_hat, u_ = carry
                (v_,) = kernel_transpose(u_hat)
                v_hat = _l2_normalize(v_)
                u_ = kernel_apply(v_hat)
                u_hat = _l2_normalize(u_)
                return (u_hat, v_hat, u_), None

            (u_hat, v_hat, u_), _ = jax.lax.scan(
                scan_body, (u_hat, v_hat, u_), None, length=self.iteration
            )
            u.value, v.value = u_hat, v_hat

        sigma = jnp.vdot(u_hat, u_)
        # Bound spectral norm by the `norm_multiplier`.
        sigma = jnp.maximum(sigma / self.norm_multiplier, 1.0)
        w_hat = w / jax.lax.stop_gradient(sigma)
        self.sow("intermediates", "w", w_hat)

        # Update params.
        params = flax.core.unfreeze(params)
        params[self.kernel_name] = w_hat
        layer_params = flax.core.freeze({"params": params})
        return self.layer.apply(layer_params, inputs)


class SpectralNormalizationConv2D(SpectralNormalization):
    __doc__ = (
        "Implements spectral normalization for Convolutional layers."
        "See `Generalizable Adversarial Training via Spectral Normalization <https://arxiv.org/abs/1811.07457>`_.\n"
        + "\n".join(SpectralNormalization.__doc__.split("\n")[1:])
    )

    kernel_apply_kwargs: Mapping[str, Any] = flax.core.FrozenDict(
        feature_group_count=1, padding="SAME", use_bias=False
    )


@dataclasses.dataclass
class WithSpectralConv2DNorm:
    """
    Attributes
    ----------
    spectral_norm_iteration: int
        The number of power iterations to estimate weight matrix's singular value.
    spectral_norm_bound: float
        Multiplicative constant to threshold the normalization.
        Usually under normalization, the singular value will converge to this value.
    """

    spectral_norm_iteration: int = 1.0
    spectral_norm_bound: float = 0.95

    def spectral_norm(self, layer: Type[nn.Module], train: bool = False) -> Callable:
        return lambda *args, **kwargs: SpectralNormalizationConv2D(
            layer=layer(*args, **kwargs),
            iteration=self.spectral_norm_iteration,
            norm_multiplier=self.spectral_norm_bound,
            update_singular_value_estimate=train,
        )


@dataclasses.dataclass
class WithSpectralNorm:
    """
    Attributes
    ----------
    spectral_norm_iteration: int
        The number of power iterations to estimate weight matrix's singular value.
    spectral_norm_bound: float
        Multiplicative constant to threshold the normalization.
        Usually under normalization, the singular value will converge to this value.
    """

    spectral_norm_iteration: int = 1.0
    spectral_norm_bound: float = 0.95

    def spectral_norm(self, layer: Type[nn.Module], train: bool = False) -> Callable:
        return lambda *args, **kwargs: SpectralNormalization(
            layer=layer(*args, **kwargs),
            iteration=self.spectral_norm_iteration,
            norm_multiplier=self.spectral_norm_bound,
            update_singular_value_estimate=train,
        )



import logging
from typing import (
    Optional,
    Union,
)

from jax import vmap
import jax.numpy as jnp

from fortuna.typing import Array


class MaxCoverageFixedPrecisionBinaryClassificationCalibrator:
    def __init__(self):
        """
        Given a binary classification framework, let us define true positive precision by
        :math:`\mathbb{P}(Y=1|f(X)\ge T_{tp})`
        and true negative precision by
        :math:`\mathbb{P}(Y=0|f(X)\le T_{tn})`,
        where :math:`T_{tp}` and :math:`T_{tn}` are two thresholds greater than :math:`\frac{1}{2}`,
        and :math:`f(X)` is a model for the probability that :math:`Y=1`.
        We further define coverage as
        :math:`\mathbb{P}(f(X)\le T_{tn} + \mathbb{P}(f(X)\ge T_{tp}`.
        Then this algorithm defines a new model
        :math:`\hat{f}(x)=(\tau_{tp}\,1[f(x)\ge \frac{1}{2}] + \tau_{tn}f(x)\,1[f(x)<\frac{1}{2}])\,f(x)`,
        and searches for the :math:`tau_{tp}\in[1, 2T_{tp}]` and :math:`tau_{tn}\in[2T_{tp}, 1]`
        that maximize the coverage while guaranteeing that true positive and negative precisions are at least
        :math:`T_{tp}` and `T_{tn}`, respectively.
        """
        self._patches = dict()

    def calibrate(
        self,
        targets: Array,
        probs: Array,
        true_positive_precision_threshold: float,
        true_negative_precision_threshold: float,
        test_probs: Optional[Array] = None,
        n_taus: int = 100,
        margin: float = 0.0,
    ) -> Union[None, Array]:
        if (
            true_negative_precision_threshold <= 0.5
            or true_positive_precision_threshold <= 0.5
        ):
            raise ValueError(
                "Both `true_negative_precision_threshold` and"
                " `true_positive_precision_threshold` must be greater than 0.5."
            )
        probs = jnp.copy(probs)
        targets = jnp.copy(targets)

        def _true_positive_objective_fn(tau: Array):
            calib_probs = jnp.clip((1 + (tau - 1) * (probs > 0.5)) * probs, a_max=1)
            b_pos_prec = calib_probs >= true_positive_precision_threshold
            prob_b_pos_prec = jnp.mean(b_pos_prec)
            pos_prec = jnp.mean(targets * b_pos_prec) / prob_b_pos_prec
            pos_cond = pos_prec >= true_positive_precision_threshold + margin
            return prob_b_pos_prec * pos_cond

        def _true_negative_objective_fn(tau: Array):
            calib_probs = (1 + (tau - 1) * (probs < 0.5)) * probs
            b_neg_prec = calib_probs <= 1 - true_negative_precision_threshold
            prob_b_neg_prec = jnp.mean(b_neg_prec)
            neg_prec = jnp.mean((1 - targets) * b_neg_prec) / prob_b_neg_prec
            neg_cond = neg_prec >= true_negative_precision_threshold + margin
            return prob_b_neg_prec * neg_cond

        taus_pos = jnp.linspace(1, 2 * true_positive_precision_threshold, n_taus)
        taus_neg = jnp.linspace(2 * (1 - true_negative_precision_threshold), 1, n_taus)[
            ::-1
        ]

        values_pos = vmap(_true_positive_objective_fn)(taus_pos)

        msg = "The {} could not be satisfied. Please consider improving the classifier or decreasing the threshold."

        if jnp.max(values_pos) == 0:
            logging.warning(msg.format("`true_positive_precision_threshold`"))
        values_neg = vmap(_true_negative_objective_fn)(taus_neg)
        if jnp.max(values_neg) == 0:
            logging.warning(msg.format("`true_negative_precision_threshold`"))

        self._patches["tau_pos"] = taus_pos[jnp.argmax(values_pos)]
        self._patches["tau_neg"] = taus_neg[jnp.argmax(values_neg)]

        if test_probs is not None:
            return self.apply_patches(test_probs)

    def apply_patches(self, probs: Array) -> Array:
        if not len(self._patches):
            logging.warning("No patches available.")
            return probs

        probs = jnp.copy(probs)
        probs = probs.at[probs > 0.5].set(
            jnp.clip(self._patches["tau_pos"] * probs[probs > 0.5], a_max=1)
        )
        probs = probs.at[probs < 0.5].set(self._patches["tau_neg"] * probs[probs < 0.5])
        return probs

    @staticmethod
    def true_positive_precision(probs: Array, targets: Array, threshold: float):
        b = probs >= threshold
        prob_b = jnp.mean(b)
        return jnp.mean(targets * b) / prob_b

    @staticmethod
    def true_negative_precision(probs: Array, targets: Array, threshold: float):
        b = probs <= 1 - threshold
        prob_b = jnp.mean(b)
        return jnp.mean((1 - targets) * b) / prob_b

    @staticmethod
    def true_positive_coverage(probs: Array, threshold: float):
        return jnp.mean(probs >= threshold)

    @staticmethod
    def true_negative_coverage(probs: Array, threshold: float):
        return jnp.mean(probs <= threshold)

    @property
    def patches(self):
        return self._patches



import pathlib
from typing import (
    Dict,
    Iterable,
    Tuple,
    Union,
)

from flax.core import FrozenDict
import jax.numpy as jnp
import numpy as np
from optax._src.base import (
    GradientTransformation,
    PyTree,
)

Params = FrozenDict[str, FrozenDict[str, PyTree]]
Mutable = FrozenDict[str, FrozenDict[str, PyTree]]
CalibParams = FrozenDict[str, PyTree]
CalibMutable = FrozenDict[str, PyTree]
OptaxOptimizer = GradientTransformation
Array = Union[jnp.ndarray, np.ndarray]
Status = Dict[str, Array]
Path = Union[str, pathlib.Path]
InputData = Union[Array, Dict[str, Array]]
Targets = Array
Batch = Tuple[InputData, Targets]
Outputs = jnp.ndarray
Uncertainties = jnp.ndarray
Predictions = jnp.ndarray
AnyKey = Union[str, int]
Shape = Union[Iterable[int], Dict[str, Iterable[int]]]



from typing import (
    List,
    Optional,
    Union,
)

from jax._src.prng import PRNGKeyArray
import jax.numpy as jnp

from fortuna.calib_model.predictive.base import Predictive
from fortuna.data.loader import InputsLoader
from fortuna.likelihood.regression import RegressionLikelihood
from fortuna.typing import Array


class RegressionPredictive(Predictive):
    def __init__(self, likelihood: RegressionLikelihood):
        super().__init__(likelihood=likelihood)

    def entropy(
        self,
        inputs_loader: InputsLoader,
        n_samples: int = 30,
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
    ) -> jnp.ndarray:
        r"""
        Estimate the predictive entropy, that is

        .. math::
            -\mathbb{E}_{Y|x, \mathcal{D}}[\log p(Y|x, \mathcal{D})],

        where:
         - :math:`x` is an observed input variable;
         - :math:`Y` is a random target variable;
         - :math:`\mathcal{D}` is the observed training data set;
         - :math:`W` denotes the random model parameters.

        Parameters
        ----------
        inputs_loader : InputsLoader
            A loader of input data points.
        n_samples : int
            Number of samples to draw for each input.
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            An estimate of the predictive entropy for each input.
        """
        state = self.state.get()
        return self.likelihood.entropy(
            params=state.params,
            inputs_loader=inputs_loader,
            mutable=state.mutable,
            distribute=distribute,
            n_target_samples=n_samples,
            rng=rng,
        )

    def quantile(
        self,
        q: Union[float, Array, List],
        inputs_loader: InputsLoader,
        n_samples: Optional[int] = 30,
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
    ) -> Union[float, jnp.ndarray]:
        r"""
        Estimate the `q`-th quantiles of the predictive probability density function.

        Parameters
        ----------
        q : Union[float, Array, List]
            Quantile or sequence of quantiles to compute. Each of these must be between 0 and 1, extremes included.
        inputs_loader : InputsLoader
            A loader of input data points.
        n_samples : int
            Number of target samples to sample for each input data point.
        rng: Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            Quantile estimate for each quantile and each input. If multiple quantiles `q` are given, the result's
            first axis is over different quantiles.
        """
        state = self.state.get()
        return self.likelihood.quantile(
            q=q,
            params=state.params,
            inputs_loader=inputs_loader,
            mutable=state.mutable,
            n_target_samples=n_samples,
            rng=rng,
            distribute=distribute,
        )

    def credible_interval(
        self,
        inputs_loader: InputsLoader,
        n_samples: int = 30,
        error: float = 0.05,
        interval_type: str = "two-tailed",
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
    ) -> jnp.ndarray:
        r"""
        Estimate credible intervals for the target variable. This is supported only if the target variable is scalar.

        Parameters
        ----------
        inputs_loader : InputsLoader
            A loader of input data points.
        n_samples: int
            Number of target samples to draw for each input.
        error: float
            The interval error. This must be a number between 0 and 1, extremes included. For example,
            `error=0.05` corresponds to a 95% level of credibility.
        interval_type: str
            The interval type. We support "two-tailed" (default), "right-tailed" and "left-tailed".
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            A credibility interval for each of the inputs.
        """
        supported_types = ["two-tailed", "right-tailed", "left-tailed"]
        if interval_type not in supported_types:
            raise ValueError(
                "`type={}` not recognised. Please choose among the following supported types: {}.".format(
                    interval_type, supported_types
                )
            )
        q = (
            jnp.array([0.5 * error, 1 - 0.5 * error])
            if interval_type == "two-tailed"
            else error
            if interval_type == "left-tailed"
            else 1 - error
        )
        qq = self.quantile(
            q=q,
            inputs_loader=inputs_loader,
            n_samples=n_samples,
            rng=rng,
            distribute=distribute,
        )
        if qq.shape[-1] != 1:
            raise ValueError(
                """Credibility intervals are only supported for scalar target variables."""
            )
        if interval_type == "two-tailed":
            lq, uq = qq.squeeze(2)
            return jnp.array(list(zip(lq, uq)))
        else:
            return qq



from typing import Union

from fortuna.prob_model.posterior.sgmcmc.base import SGMCMCPosteriorApproximator
from fortuna.prob_model.posterior.sgmcmc.sghmc import SGHMC_NAME
from fortuna.prob_model.posterior.sgmcmc.sgmcmc_preconditioner import (
    Preconditioner,
    identity_preconditioner,
)
from fortuna.prob_model.posterior.sgmcmc.sgmcmc_step_schedule import (
    StepSchedule,
    constant_schedule,
)


class SGHMCPosteriorApproximator(SGMCMCPosteriorApproximator):
    def __init__(
        self,
        n_samples: int = 10,
        n_thinning: int = 1,
        burnin_length: int = 1000,
        momentum_decay: float = 0.01,
        step_schedule: Union[StepSchedule, float] = 1e-5,
        preconditioner: Preconditioner = identity_preconditioner(),
    ) -> None:
        """
        SGHMC posterior approximator. It is responsible to define how the posterior distribution is approximated.

        The total number of available posterior samples depends on the number of training steps, `burnin_length`,
        and `n_thinning` parameters:

        `n_available_samples` = (`n_training_steps` - `burnin_length`) % `n_thinning`

        Setting the desired number of samples `n_samples` larger than `n_available_samples` will result in an
        exception.

        Parameters
        ----------
        n_samples: int
            The desired number of the posterior samples.
        n_thinning: int
            If `n_thinning` > 1, keep only each `n_thinning` sample during the sampling phase.
        burnin_length: int
            Length of the initial burn-in phase, in steps.
        momentum_decay: float
            The "friction" term that counters the noise of stochastic gradient estimates. Setting this argument to zero recovers the overamped Langevin dynamics.
        step_schedule: Union[StepSchedule, float]
            Either a constant `float` step size or a schedule function.
        preconditioner: Preconditioner
            A `Preconditioner` instance that preconditions the approximator with information about the posterior distribution, if available.

        """
        super().__init__(
            n_samples=n_samples,
            n_thinning=n_thinning,
            preconditioner=preconditioner,
        )
        if isinstance(step_schedule, float):
            step_schedule = constant_schedule(step_schedule)
        elif not callable(step_schedule):
            raise ValueError(f"`step_schedule` must be a a callable function.")
        self.burnin_length = burnin_length
        self.momentum_decay = momentum_decay
        self.step_schedule = step_schedule

    def __str__(self) -> str:
        return SGHMC_NAME



from __future__ import annotations

import logging
from typing import Optional

from flax.core import FrozenDict
from jax import random
from jax._src.prng import PRNGKeyArray
from jax.flatten_util import ravel_pytree
import jax.numpy as jnp

from fortuna.data.loader import (
    DataLoader,
    InputsLoader,
)
from fortuna.prob_model.fit_config.base import FitConfig
from fortuna.prob_model.joint.base import Joint
from fortuna.prob_model.joint.state import JointState
from fortuna.prob_model.posterior.base import Posterior
from fortuna.prob_model.posterior.map.map_state import MAPState
from fortuna.prob_model.posterior.posterior_state_repository import (
    PosteriorStateRepository,
)
from fortuna.prob_model.posterior.run_preliminary_map import run_preliminary_map
from fortuna.prob_model.posterior.swag import SWAG_NAME
from fortuna.prob_model.posterior.swag.swag_approximator import (
    SWAGPosteriorApproximator,
)
from fortuna.prob_model.posterior.swag.swag_state import SWAGState
from fortuna.prob_model.posterior.swag.swag_trainer import (
    JittedSWAGTrainer,
    MultiDeviceSWAGTrainer,
    SWAGTrainer,
)
from fortuna.typing import (
    Array,
    Status,
)
from fortuna.utils.device import select_trainer_given_devices
from fortuna.utils.freeze import get_trainable_paths
from fortuna.utils.nested_dicts import (
    nested_get,
    nested_set,
)
from fortuna.utils.strings import decode_encoded_tuple_of_lists_of_strings_to_array


class SWAGPosterior(Posterior):
    def __init__(self, joint: Joint, posterior_approximator: SWAGPosteriorApproximator):
        """
        SWAG approximate posterior class.

        Parameters
        ----------
        joint: Joint
            A joint distribution object.
        posterior_approximator: SWAGPosteriorApproximator
            A SWAG posterior approximator.
        """
        super().__init__(joint=joint, posterior_approximator=posterior_approximator)

    def __str__(self):
        return SWAG_NAME

    def fit(
        self,
        train_data_loader: DataLoader,
        val_data_loader: Optional[DataLoader] = None,
        fit_config: FitConfig = FitConfig(),
        map_fit_config: Optional[FitConfig] = None,
        **kwargs,
    ) -> Status:
        super()._checks_on_fit_start(fit_config, map_fit_config)

        if self.posterior_approximator.rank < 2:
            raise ValueError("`rank` must be at least 2.")
        if fit_config.optimizer.n_epochs <= self.posterior_approximator.rank:
            raise ValueError(
                """Not enough SWAG epochs to obtain `rank={}`. Please either increase `n_swag_epochs` or
            decrease `rank`.""".format(
                    self.posterior_approximator.rank
                )
            )
        if (
            fit_config.monitor.early_stopping_patience
            and fit_config.monitor.early_stopping_patience > 0
        ):
            logging.warning(
                f"""It seems you are trying to enable early stopping, since
            `fit_config.monitor.early_stopping_patience={fit_config.monitor.early_stopping_patience}`. We do not
            support early stopping in SWAG, since we implement it as a post-processing step of MAP. If your intention
            was rather to enable early stopping in MAP, please configure `map_fit_config` accordingly."""
            )

        status = dict()

        if super()._is_state_available_somewhere(fit_config):
            state = super()._restore_state_from_somewhere(
                fit_config=fit_config, allowed_states=(MAPState, SWAGState)
            )

        elif super()._should_run_preliminary_map(fit_config, map_fit_config):
            state, status["map"] = run_preliminary_map(
                joint=self.joint,
                train_data_loader=train_data_loader,
                val_data_loader=val_data_loader,
                map_fit_config=map_fit_config,
                rng=self.rng,
                **kwargs,
            )
        else:
            raise ValueError(
                "The SWAG approximation must start from a preliminary run of MAP or an existing "
                "checkpoint or state. Please configure `map_fit_config`, or "
                "`fit_config.checkpointer.restore_checkpoint_path`, "
                "or `fit_config.checkpointer.start_from_current_state`."
            )

        state = SWAGState.convert_from_map_state(
            map_state=state,
            optimizer=fit_config.optimizer.method,
        )

        state = super()._freeze_optimizer_in_state(state, fit_config)

        if fit_config.optimizer.freeze_fun is not None:
            which_params = get_trainable_paths(
                state.params, fit_config.optimizer.freeze_fun
            )
        else:
            which_params = None

        trainer_cls = select_trainer_given_devices(
            devices=fit_config.processor.devices,
            base_trainer_cls=SWAGTrainer,
            jitted_trainer_cls=JittedSWAGTrainer,
            multi_device_trainer_cls=MultiDeviceSWAGTrainer,
            disable_jit=fit_config.processor.disable_jit,
        )
        trainer = trainer_cls(
            predict_fn=self.joint.likelihood.prob_output_layer.predict,
            save_checkpoint_dir=fit_config.checkpointer.save_checkpoint_dir,
            save_every_n_steps=fit_config.checkpointer.save_every_n_steps,
            keep_top_n_checkpoints=fit_config.checkpointer.keep_top_n_checkpoints,
            disable_training_metrics_computation=fit_config.monitor.disable_training_metrics_computation,
            eval_every_n_epochs=fit_config.monitor.eval_every_n_epochs,
            early_stopping_verbose=False,
            freeze_fun=fit_config.optimizer.freeze_fun,
            which_params=which_params,
        )

        kwargs = dict(rank=self.posterior_approximator.rank)
        logging.info("Run SWAG.")
        state, status["swag"] = trainer.train(
            rng=self.rng.get(),
            state=state,
            loss_fun=self.joint._batched_negative_log_joint_prob,
            training_dataloader=train_data_loader,
            training_dataset_size=train_data_loader.size,
            n_epochs=fit_config.optimizer.n_epochs,
            metrics=fit_config.monitor.metrics,
            validation_dataloader=val_data_loader,
            validation_dataset_size=val_data_loader.size
            if val_data_loader is not None
            else None,
            verbose=fit_config.monitor.verbose,
            callbacks=fit_config.callbacks,
            **kwargs,
        )

        self.state = PosteriorStateRepository(
            fit_config.checkpointer.save_checkpoint_dir
            if fit_config.checkpointer.dump_state is True
            else None
        )
        self.state.put(state, keep=fit_config.checkpointer.keep_top_n_checkpoints)
        logging.info("Fit completed.")
        return status

    def sample(
        self,
        rng: Optional[PRNGKeyArray] = None,
        inputs_loader: Optional[InputsLoader] = None,
        inputs: Optional[Array] = None,
        **kwargs,
    ) -> JointState:
        """
        Sample from the posterior distribution.

        Parameters
        ----------
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        inputs_loader: Optional[InputsLoader]
            Input data loader. This or `inputs` is required if the posterior state includes mutable objects.
        inputs: Optional[Array]
            Input variables. This or `inputs_loader` is required if the posterior state includes mutable objects.

        Returns
        -------
        JointState
            A sample from the posterior distribution.
        """
        if rng is None:
            rng = self.rng.get()
        state = self.state.get()
        if state.mutable is not None and inputs_loader is None and inputs is None:
            raise ValueError(
                "The posterior state contains mutable objects. Please pass `inputs_loader` or `inputs`."
            )

        n_params = len(state.mean)
        rank = state.dev.shape[-1]
        which_params = decode_encoded_tuple_of_lists_of_strings_to_array(
            state._encoded_which_params
        )

        unravel = ravel_pytree(
            state.params
            if which_params is None
            else [nested_get(state.params, path) for path in which_params]
        )[1]

        coeff1 = 1 / jnp.sqrt(2)
        coeff2 = coeff1 / jnp.sqrt(rank)

        rng, key1, key2 = random.split(rng, 3)
        z1 = random.normal(key1, shape=(n_params,))
        z2 = random.normal(key2, shape=(rank,))
        if which_params is None:
            state = state.replace(
                params=self._get_sample(
                    mean=state.mean,
                    std=state.std,
                    dev=state.dev,
                    z1=z1,
                    z2=z2,
                    coeff1=coeff1,
                    coeff2=coeff2,
                    unravel=unravel,
                )
            )
        else:
            state = state.replace(
                params=FrozenDict(
                    nested_set(
                        d=state.params.unfreeze(),
                        key_paths=which_params,
                        objs=tuple(
                            self._get_sample(
                                mean=state.mean,
                                std=state.std,
                                dev=state.dev,
                                z1=z1,
                                z2=z2,
                                coeff1=coeff1,
                                coeff2=coeff2,
                                unravel=unravel,
                            )
                        ),
                    )
                )
            )

        if state.mutable:
            if inputs_loader is not None:
                for batch_inputs in inputs_loader:
                    state = state.replace(
                        mutable=self.joint.likelihood.model_manager.apply(
                            state.params,
                            batch_inputs,
                            mutable=state.mutable,
                            train=True,
                            rng=rng,
                        )[1]["mutable"]
                    )
            else:
                state = state.replace(
                    mutable=self.joint.likelihood.model_manager.apply(
                        state.params, inputs, mutable=state.mutable, train=True, rng=rng
                    )[1]["mutable"]
                )

        return JointState(
            params=state.params,
            mutable=state.mutable,
            calib_params=state.calib_params,
            calib_mutable=state.calib_mutable,
        )

    def _get_sample(self, mean, std, dev, z1, z2, coeff1, coeff2, unravel):
        return unravel(mean + coeff1 * std * z1 + coeff2 * jnp.matmul(dev, z2))

    def _get_mean_std_dev(self, state: SWAGState) -> SWAGState:
        var = state._mean_squared_rav_params - state._mean_rav_params**2
        var = jnp.maximum(var, 0.0)
        return state.update(
            dict(
                mean=state._mean_rav_params,
                std=jnp.sqrt(var),
                dev=state._deviation_rav_params,
            )
        )



from __future__ import annotations

from copy import deepcopy
from itertools import zip_longest
from typing import (
    Iterable,
    Optional,
    Tuple,
    Union,
)

import numpy as np

from fortuna.typing import (
    Array,
    Batch,
    InputData,
    Targets,
)


class IterableData:
    def __init__(
        self,
        generator,
    ):
        self._g = generator

    def __iter__(self) -> Iterable[Union[Batch, InputData, Targets]]:
        yield from iter(self._g())

    @classmethod
    def from_iterable(cls, iterable) -> IterableData:
        def _inner():
            yield from iterable

        return cls(_inner)

    @classmethod
    def from_callable(cls, fn) -> IterableData:
        def _inner():
            yield from fn()

        return cls(_inner)

    @classmethod
    def from_tf_dataloader(cls, tf_dataloader) -> IterableData:
        def _inner():
            for batch_inputs, batch_targets in tf_dataloader:
                if not isinstance(batch_inputs, dict):
                    batch_inputs = batch_inputs.numpy()
                else:
                    batch_inputs = {k: v.numpy() for k, v in batch_inputs.items()}
                batch_targets = batch_targets.numpy()
                yield batch_inputs, batch_targets

        return cls(_inner)

    @classmethod
    def from_torch_dataloader(cls, torch_dataloader) -> IterableData:
        def _inner():
            for batch_inputs, batch_targets in torch_dataloader:
                if not isinstance(batch_inputs, dict):
                    batch_inputs = batch_inputs.numpy()
                else:
                    batch_inputs = {k: v.numpy() for k, v in batch_inputs.items()}
                batch_targets = batch_targets.numpy()
                yield batch_inputs, batch_targets

        return cls(_inner)

    @classmethod
    def from_array_data(
        cls,
        data: Array,
        batch_size: Optional[int] = None,
        shuffle: bool = False,
        prefetch: bool = False,
        seed: Optional[int] = 0,
    ):
        def _inner():
            if shuffle:
                rng = np.random.default_rng(seed)
                perm = rng.choice(data.shape[0], data.shape[0], replace=False)
            if batch_size is None:
                yield data
            else:
                batches = np.split(
                    data[perm] if shuffle else data,
                    np.arange(batch_size, data.shape[0], batch_size),
                    axis=0,
                )

                def make_gen():
                    for batch in batches:
                        yield batch

                yield from PrefetchedGenerator(make_gen()) if prefetch else make_gen()

        return cls(_inner)

    @classmethod
    def from_batch_array_data(
        cls,
        data: Tuple[Array, Array],
        batch_size: Optional[int] = None,
        shuffle: bool = False,
        prefetch: bool = False,
        seed: Optional[int] = 0,
    ):
        def _inner():
            if shuffle:
                rng = np.random.default_rng(seed)
                perm = rng.choice(data[0].shape[0], data[0].shape[0], replace=False)
            if batch_size is None:
                yield data
            else:
                x_batches = np.split(
                    data[0][perm] if shuffle else data[0],
                    np.arange(batch_size, data[0].shape[0], batch_size),
                    axis=0,
                )
                y_batches = np.split(
                    data[1][perm] if shuffle else data[1],
                    np.arange(batch_size, data[1].shape[0], batch_size),
                    axis=0,
                )

                def make_gen():
                    for x_batch, y_batch in zip(x_batches, y_batches):
                        yield x_batch, y_batch

                yield from PrefetchedGenerator(make_gen()) if prefetch else make_gen()

        return cls(_inner)

    @classmethod
    def data_loader_to_inputs_iterable(cls, data_loader) -> IterableData:
        def _inner():
            for inputs, _ in data_loader:
                yield inputs

        return cls(_inner)

    @classmethod
    def data_loader_to_targets_iterable(cls, data_loader) -> IterableData:
        def _inner():
            for _, targets in data_loader:
                yield targets

        return cls(_inner)

    @classmethod
    def inputs_loaders_to_batch_iterable(
        cls, inputs_loaders, targets, how
    ) -> IterableData:
        if how == "interpose":

            def _inner():
                for all_inputs in zip_longest(*inputs_loaders):
                    _targets = [
                        target * np.ones(inputs.shape[0], dtype="int32")
                        for inputs, target in zip(all_inputs, targets)
                        if inputs is not None
                    ]
                    all_inputs = [inputs for inputs in all_inputs if inputs is not None]
                    yield np.concatenate(all_inputs), np.concatenate(_targets)

        elif how == "concatenate":

            def _inner():
                for inputs_loader, target in zip(inputs_loaders, targets):
                    for inputs in inputs_loader:
                        yield inputs, target * np.ones(inputs.shape[0], dtype="int32")

        else:
            hows = ["interpose", "concatenate"]
            raise ValueError(
                f"`how={how} not recognized. Please choose among the following options: {hows}."
            )
        return cls(_inner)

    @classmethod
    def transform_data_loader(cls, loader, transform_fn, status) -> IterableData:
        def _inner():
            _status = deepcopy(status)
            for inputs, targets in loader:
                inputs, targets, _status = transform_fn(inputs, targets, _status)
                if (
                    inputs is not None
                    and targets is not None
                    and len(inputs) > 0
                    and len(targets) > 0
                ):
                    if inputs.shape[0] != targets.shape[0]:
                        raise ValueError(
                            "The first dimension of transformed inputs and targets must be the same, but "
                            f"{inputs.shape[0]} and {targets.shape[0]} were found."
                        )
                    yield inputs, targets

        return cls(_inner)

    @classmethod
    def transform_inputs_or_targets_loader(
        cls, loader, transform_fn, status
    ) -> IterableData:
        def _inner():
            _status = deepcopy(status)
            for data in loader:
                data, _status = transform_fn(data, _status)
                if data is not None and len(data) > 0:
                    yield data

        return cls(_inner)


class PrefetchedGenerator:
    def __init__(self, generator):
        self._batch = generator.__next__()
        self._generator = generator
        self._ready = True

    def __iter__(self):
        return self

    def __next__(self):
        if not self._ready:
            self._prefetch()
        self._ready = False
        return self._batch

    def _prefetch(self):
        if not self._ready:
            self._batch = self._generator.__next__()
            self._ready = True



from fortuna.prob_model.posterior.base import PosteriorApproximator
from fortuna.prob_model.posterior.sgmcmc.sgmcmc_preconditioner import (
    Preconditioner,
    identity_preconditioner,
)


class SGMCMCPosteriorApproximator(PosteriorApproximator):
    def __init__(
        self,
        n_samples: int = 10,
        n_thinning: int = 1,
        preconditioner: Preconditioner = identity_preconditioner(),
    ) -> None:
        """
        SGMCMC posterior approximator. It is responsible to define how the posterior distribution is approximated.

        Parameters
        ----------
        n_samples: int
            The desired number of the posterior samples.
        n_thinning: int
            If `n_thinning` > 1, keep only each `n_thinning` sample during the sampling phase.
        preconditioner: Preconditioner
            A `Preconditioner` instance that preconditions the approximator with information about the posterior distribution, if available.

        """
        self.n_samples = n_samples
        self.n_thinning = n_thinning
        self.preconditioner = preconditioner

    def __str__(self) -> str:
        raise NotImplementedError



from typing import Optional

from fortuna.training.callback import Callback
from fortuna.training.train_state import TrainState
from fortuna.training.train_state_repository import TrainStateRepository
from fortuna.training.trainer import TrainerABC


class SGMCMCSamplingCallback(Callback):
    def __init__(
        self,
        trainer: TrainerABC,
        state_repository: TrainStateRepository,
        keep_top_n_checkpoints: int,
    ):
        """
        Sampling callback that collects samples from the MCMC chain.

        Parameters
        ----------
        trainer: TrainerABC
            An instance of the trainer class.
        state_repository: TrainStateRepository
            An instance of the state repository.
        keep_top_n_checkpoints: int
            Number of past checkpoint files to keep.
        """
        self._trainer = trainer
        self._state_repository = state_repository
        self._keep_top_n_checkpoints = keep_top_n_checkpoints

        self._current_step = 0
        self._samples_count = 0

    def _do_sample(self, current_step, samples_count):
        raise NotImplementedError

    def training_step_end(self, state: TrainState) -> TrainState:
        self._current_step += 1

        if self._do_sample(self._current_step, self._samples_count):
            self._state_repository.put(
                state=self._trainer._sync_state(state),
                i=self._samples_count,
                keep=self._keep_top_n_checkpoints,
            )
            self._samples_count += 1

        return state



from jax import random
from jax._src.prng import PRNGKeyArray
from jax.tree_util import (
    tree_map,
    tree_structure,
    tree_unflatten,
)
from optax._src.base import PyTree


def generate_rng_like_tree(rng, target: PyTree):
    treedef = tree_structure(target)
    keys = random.split(rng, treedef.num_leaves)
    return tree_unflatten(treedef, keys)


def generate_random_normal_like_tree(rng, target: PyTree):
    keys = generate_rng_like_tree(rng, target)
    return tree_map(
        lambda l, k: random.normal(k, l.shape, l.dtype),
        target,
        keys,
    )


class RandomNumberGenerator:
    def __init__(self, seed: int):
        """
        A random number generator object.

        Parameters
        ----------
        seed : int
            A random seed.
        """
        self._rng = random.PRNGKey(seed)

    def get(self) -> PRNGKeyArray:
        """
        Get the internal random number generator key. Whenever this function is called, the random number generator
        key is updated.

        Returns
        -------
        PRNGKeyArray
            A random number generator key.
        """
        self._rng = random.split(self._rng)[0]
        return self._rng


class WithRNG:
    @property
    def rng(self) -> RandomNumberGenerator:
        """
        Invoke the random number generator object.

        Returns
        -------
        The random number generator object.
        """
        return self._rng

    @rng.setter
    def rng(self, rng: RandomNumberGenerator):
        """
        Set a random number generator object.

        Parameters
        ----------
        rng : RandomNumberGenerator
            A random number generator object.
        """
        self._rng = rng



from typing import (
    Dict,
    Optional,
    Tuple,
    Union,
)

import jax.nn
import jax.numpy as jnp

from fortuna.data.loader import TargetsLoader
from fortuna.plot import plot_reliability_diagram
from fortuna.typing import Array


def accuracy(preds: Array, targets: Array) -> jnp.ndarray:
    """
    Compute the accuracy given predictions and target variables.

    Parameters
    ----------
    preds: Array
        A one-dimensional array of predictions over the data points.
    targets: Array
        A one-dimensional array of target variables.

    Returns
    -------
    jnp.ndarray
        The computed accuracy.
    """
    if preds.ndim > 1:
        raise ValueError(
            """`preds` must be a one-dimensional array of predicted classes."""
        )
    if targets.ndim > 1:
        raise ValueError(
            """`targets` must be a one-dimensional array of target classes."""
        )
    return jnp.mean(preds == targets)


def compute_counts_confs_accs(
    preds: Array,
    probs: Array,
    targets: Array,
    plot: bool = False,
    plot_options: Optional[Dict] = None,
) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]:
    """
    Bin the confidence scores (maximum probability) and for each of them compute:
        - the number of inputs;
        - the average confidence score for each bin;
        - the average accuracy over each bin.

    Parameters
    ----------
    preds: Array
        A one-dimensional array of predictions over the data points.
    probs: Array
        A two-dimensional array of class probabilities for each data point.
    targets: Array
        A one-dimensional array of target variables.
    plot: bool
        Whether to plot a reliability diagram.
    plot_options: dict
        Options for the reliability diagram plot; see :func:`~fortuna.plot.plot_reliability_diagram`.

    Returns
    -------
    Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]
        Number of inputs per bin, average confidence score per bin and average accuracy per bin.
    """
    if probs.ndim == 2:
        probs = probs.max(-1)
    thresholds = jnp.linspace(1 / len(probs), 1, 10)
    probs = jnp.array(probs)
    indices = [jnp.where(probs <= thresholds[0])[0]]
    indices += [
        jnp.where((probs <= thresholds[i]) & (probs > thresholds[i - 1]))[0]
        for i in range(1, len(thresholds))
    ]
    counts = jnp.array([len(idx) for idx in indices])

    diff = targets - preds
    accs = jnp.array([jnp.nan_to_num(jnp.mean(diff[idx] == 0)) for idx in indices])
    confs = jnp.array([jnp.nan_to_num(jnp.mean(probs[idx])) for idx in indices])

    if plot:
        idx = confs != 0
        if plot_options is None:
            plot_options = dict()
        plot_reliability_diagram(accs[idx], confs[idx], **plot_options)

    return counts, confs, accs


def expected_calibration_error(
    preds: Array,
    probs: Array,
    targets: Array,
    plot: bool = False,
    plot_options: Optional[Dict] = None,
) -> jnp.ndarray:
    """
    Compute the Expected Calibration Error (ECE)
    (see `Naeini et al., 2015 <https://people.cs.pitt.edu/~milos/research/AAAI_Calibration.pdf>`__ and
    `Guo et al., 2017 <http://proceedings.mlr.press/v70/guo17a/guo17a.pdf>`__). Optionally, plot and save a reliability
    diagram.

    Parameters
    ----------
    preds: Array
        A one-dimensional array of predictions over the data points.
    probs: Array
        A two-dimensional array of class probabilities for each data point.
    targets: Array
        A one-dimensional array of target variables.
    plot: bool
        Whether to plot a reliability diagram.
    plot_options: dict
        Options for the reliability diagram plot; see :func:`~fortuna.plot.plot_reliability_diagram`.

    Returns
    -------
    jnp.ndarray
        The value of the ECE.
    """
    counts, confs, accs = compute_counts_confs_accs(
        preds, probs, targets, plot, plot_options
    )
    ece = jnp.sum(counts * jnp.abs(accs - confs)) / preds.shape[0]
    return ece


def ece(
    preds: Array,
    probs: Array,
    targets: Array,
    plot: bool = False,
    plot_options: Optional[Dict] = None,
) -> jnp.ndarray:
    """See :func:`.expected_calibration_error`."""
    return expected_calibration_error(preds, probs, targets, plot, plot_options)


def maximum_calibration_error(
    preds: Array,
    probs: Array,
    targets: Array,
    plot: bool = False,
    plot_options: Optional[Dict] = None,
) -> jnp.ndarray:
    """
    Compute the Maximum Calibration Error (MCE)
    (see `Naeini et al., 2015 <https://people.cs.pitt.edu/~milos/research/AAAI_Calibration.pdf>`__). Optionally, plot
    and save a reliability diagram.

    Parameters
    ----------
    preds: Array
        A one-dimensional array of predictions over the data points.
    probs: Array
        A two-dimensional array of class probabilities for each data point.
    targets: Array
        A one-dimensional array of target variables.
    plot: bool
        Whether to plot a reliability diagram.
    plot_options: dict
        Options for the reliability diagram plot; see :func:`~fortuna.plot.plot_reliability_diagram`.

    Returns
    -------
    jnp.ndarray
        The value of the MCE.
    """
    counts, confs, accs = compute_counts_confs_accs(
        preds, probs, targets, plot, plot_options
    )
    mce = jnp.max(counts * jnp.abs(accs - confs))
    return mce


def mce(
    preds: Array,
    probs: Array,
    targets: Array,
    plot: bool = False,
    plot_options: Optional[Dict] = None,
) -> jnp.ndarray:
    """See :func:`.maximum_calibration_error`."""
    return maximum_calibration_error(preds, probs, targets, plot, plot_options)


def brier_score(probs: Array, targets: Union[TargetsLoader, Array]) -> jnp.ndarray:
    """
    Brier score (see `Brier, 1950 <https://web.archive.org/web/20171023012737/
    https://docs.lib.noaa.gov/rescue/mwr/078/mwr-078-01-0001.pdf>`__). This can be used for both binary and multi-class
    classification.

    Parameters
    ----------
    probs: Array
        A one- or two-dimensional array of class probabilities for each data point.
    targets: Array
        A one-dimensional array of target variables.

    Returns
    -------
    jnp.ndarray
        The Brier score.
    """
    if probs.ndim > 2:
        raise ValueError("`probs` can be at most 2 dimensional.")
    if type(targets) == TargetsLoader:
        targets = targets.to_array_targets()
    if probs.ndim > 1:
        targets = jax.nn.one_hot(targets, probs.shape[-1])
        return jnp.mean(jnp.sum((probs - targets) ** 2, axis=-1))
    return jnp.mean((probs - targets) ** 2)



import abc
import collections
from functools import partial
import logging
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)

from flax import jax_utils
from flax.core import FrozenDict
from flax.training.common_utils import stack_forest
import jax
from jax import (
    lax,
    random,
    value_and_grad,
)
from jax._src.prng import PRNGKeyArray
import jax.numpy as jnp
from jax.tree_util import tree_map
from optax._src.base import PyTree
from tqdm import trange
from tqdm.std import tqdm as TqdmDecorator

from fortuna.data.loader import DataLoader
from fortuna.training.callback import Callback
from fortuna.training.mixin import (
    InputValidatorMixin,
    WithCheckpointingMixin,
    WithEarlyStoppingMixin,
)
from fortuna.training.train_state import TrainState
from fortuna.typing import (
    AnyKey,
    Array,
    Batch,
    CalibMutable,
    CalibParams,
    Mutable,
    Params,
    Path,
    Status,
)
from fortuna.utils.builtins import HashableMixin
from fortuna.utils.freeze import (
    get_frozen_paths,
    get_trainable_paths,
)
from fortuna.utils.nested_dicts import (
    nested_get,
    nested_set,
    nested_update,
)
from fortuna.utils.training import clip_grandients_by_norm


class TrainerABC(
    HashableMixin,
    WithCheckpointingMixin,
    WithEarlyStoppingMixin,
    InputValidatorMixin,
    metaclass=abc.ABCMeta,
):
    def __init__(
        self,
        *args,
        predict_fn: Callable[[jnp.ndarray], jnp.ndarray],
        uncertainty_fn: Optional[Callable[[jnp.ndarray], jnp.ndarray]] = None,
        save_checkpoint_dir: Optional[Path] = None,
        save_every_n_steps: Optional[int] = None,
        keep_top_n_checkpoints: int = 2,
        disable_training_metrics_computation: bool = False,
        eval_every_n_epochs: int = 1,
        freeze_fun: Optional[Callable[[Tuple[AnyKey, ...], Array], str]] = None,
        **kwargs,
    ):
        super(TrainerABC, self).__init__(*args, **kwargs)
        self.predict_fn = predict_fn
        self.uncertainty_fn = uncertainty_fn
        self.save_checkpoint_dir = save_checkpoint_dir
        self.save_every_n_steps = save_every_n_steps
        self.keep_top_n_checkpoints = keep_top_n_checkpoints
        self.disable_training_metrics_computation = disable_training_metrics_computation
        self.eval_every_n_epochs = eval_every_n_epochs
        self._global_training_step = 0
        self._unravel = None
        self.multi_device = False
        self.freeze_fun = freeze_fun

    @abc.abstractmethod
    def training_loss_step(
        self,
        loss_fun: Callable[[Any], Union[float, Tuple[float, dict]]],
        params: Params,
        batch: Batch,
        mutable: Mutable,
        rng: PRNGKeyArray,
        n_data: int,
        unravel: Optional[Callable[[any], PyTree]] = None,
        calib_params: Optional[CalibParams] = None,
        calib_mutable: Optional[CalibMutable] = None,
        kwargs: FrozenDict[str, Any] = FrozenDict(),
    ) -> Tuple[jnp.ndarray, Dict[str, Any]]:
        pass

    def training_step(
        self,
        state: TrainState,
        batch: Batch,
        loss_fun: Callable,
        rng: PRNGKeyArray,
        n_data: int,
        unravel: Optional[Callable[[any], PyTree]] = None,
        kwargs: FrozenDict[str, Any] = FrozenDict(),
    ) -> Tuple[TrainState, Dict[str, Any]]:
        if state.dynamic_scale is None:
            value_and_grad_fn = value_and_grad
        else:
            value_and_grad_fn = state.dynamic_scale.value_and_grad

        forward_fun = lambda params: self.training_loss_step(
            loss_fun,
            params,
            batch,
            state.mutable,
            rng,
            n_data,
            unravel,
            state.calib_params,
            state.calib_mutable,
            kwargs,
        )
        if self.freeze_fun is not None:
            frozen_forward_fun = lambda trainable_params: forward_fun(
                params=self._get_all_params(
                    state=state, trainable_params=trainable_params
                )
            )

        grad_fn = value_and_grad_fn(
            forward_fun if self.freeze_fun is None else frozen_forward_fun,
            has_aux=True,
        )
        outputs = grad_fn(state.params)
        if state.dynamic_scale is None:
            (loss, aux), grad = outputs
            loss = self._sync_array(loss)
        else:
            # dynamic_scale.value_and_grad takes care of averaging gradients and loss across replicas (no need to call sync)
            dynamic_scale, is_fin, aux, grad = outputs
            loss, aux = aux

        max_grad_norm = kwargs.get("max_grad_norm", 0.0) or 0.0
        gradient_accumulation_steps = (
            kwargs.get("gradient_accumulation_steps", 0.0) or 0.0
        )
        if gradient_accumulation_steps > 1:
            new_state = self.params_update_with_grad_accumulation(
                state, grad, aux, gradient_accumulation_steps, max_grad_norm
            )
        else:
            grad = self._sync_array(grad) if state.dynamic_scale is None else grad
            if max_grad_norm > 0:
                grad = clip_grandients_by_norm(grad, max_grad_norm)
            new_state = state.apply_gradients(grads=grad, mutable=aux["mutable"])

        if new_state.dynamic_scale is not None:
            # if is_fin is False the gradients contain Inf/NaNs and optimizer state and
            # params should be restored (= skip this step).
            new_state = new_state.replace(
                opt_state=jax.tree_util.tree_map(
                    partial(jnp.where, is_fin), new_state.opt_state, state.opt_state
                ),
                params=jax.tree_util.tree_map(
                    partial(jnp.where, is_fin), new_state.params, state.params
                ),
                dynamic_scale=dynamic_scale,
            )
        return (
            new_state,
            {
                "loss": loss,
                "outputs": aux["outputs"],
                "logging_kwargs": aux["logging_kwargs"],
            },
        )

    def params_update_with_grad_accumulation(
        self,
        state: TrainState,
        grad: jnp.ndarray,
        aux: Dict[str, Any],
        gradient_accumulation_steps: int,
        max_grad_norm: float,
    ) -> TrainState:
        if state.grad_accumulated is not None:
            grad_acc = tree_map(lambda x, y: x + y, grad, state.grad_accumulated)
        else:
            # grad_accumulated is not initialized yet, will do no now
            grad_acc = grad

        def update_fn(state):
            grad = tree_map(lambda x: x / gradient_accumulation_steps, grad_acc)
            grad = self._sync_array(grad) if state.dynamic_scale is None else grad
            if max_grad_norm > 0:
                grad = clip_grandients_by_norm(grad, max_grad_norm)
            return state.apply_gradients(
                grads=grad,
                grad_accumulated=jax.tree_map(jnp.zeros_like, grad),
                mutable=aux["mutable"],
            )

        new_state = jax.lax.cond(
            (state.step + 1) % gradient_accumulation_steps == 0,
            lambda state: update_fn(state),
            lambda state: state.replace(
                grad_accumulated=grad_acc, step=state.step + 1, mutable=aux["mutable"]
            ),
            state,
        )
        return new_state

    @abc.abstractmethod
    def validation_step(
        self,
        state: TrainState,
        batch: Batch,
        loss_fun: Callable,
        rng: PRNGKeyArray,
        n_data: int,
        metrics: Optional[Tuple[Callable[[jnp.ndarray, Array], float], ...]] = None,
        unravel: Optional[Callable[[any], PyTree]] = None,
        kwargs: FrozenDict[str, Any] = FrozenDict(),
    ) -> Dict[str, jnp.ndarray]:
        pass

    def training_step_end(
        self,
        current_epoch: int,
        state: TrainState,
        aux: Dict[str, Any],
        batch: Batch,
        metrics: Optional[
            Tuple[
                Union[
                    Callable[[jnp.ndarray, Array], float],
                    Callable[[jnp.ndarray, jnp.ndarray, Array], float],
                ],
                ...,
            ]
        ],
        callbacks: Optional[List[Callback]] = None,
        kwargs: FrozenDict[str, Any] = FrozenDict(),
    ) -> Tuple[TrainState, Dict[str, jnp.ndarray]]:
        if (
            self.save_checkpoint_dir is not None
            and self.save_every_n_steps is not None
            and self.save_every_n_steps > 0
            and self._global_training_step >= self.save_every_n_steps
            and self._global_training_step % self.save_every_n_steps == 0
        ):
            self.save_checkpoint(
                state, self.save_checkpoint_dir, keep=self.keep_top_n_checkpoints
            )
        training_losses_and_metrics = {"loss": aux["loss"]}

        if aux["logging_kwargs"] is not None:
            for k, v in aux["logging_kwargs"].items():
                training_losses_and_metrics[k] = v

        if not self.disable_training_metrics_computation and metrics is not None:
            preds = self.predict_fn(aux["outputs"], train=True)
            if self.uncertainty_fn is not None:
                uncertainties = self.uncertainty_fn(aux["outputs"])
                if self.multi_device:
                    training_batch_metrics = self.compute_metrics(
                        preds.reshape(
                            (preds.shape[0] * preds.shape[1],) + preds.shape[2:]
                        ),
                        batch[1].reshape(
                            (batch[1].shape[0] * batch[1].shape[1],)
                            + batch[1].shape[2:]
                        ),
                        metrics,
                        uncertainties.reshape(
                            (uncertainties.shape[0] * uncertainties.shape[1],)
                            + uncertainties.shape[2:]
                        ),
                    )
                else:
                    training_batch_metrics = self.compute_metrics(
                        preds, batch[1], metrics, uncertainties
                    )
            else:
                if self.multi_device:
                    training_batch_metrics = self.compute_metrics(
                        preds=preds.reshape(
                            (preds.shape[0] * preds.shape[1],) + preds.shape[2:]
                        ),
                        targets=batch[1].reshape(
                            (batch[1].shape[0] * batch[1].shape[1],)
                            + batch[1].shape[2:]
                        ),
                        metrics=metrics,
                    )
                else:
                    training_batch_metrics = self.compute_metrics(
                        preds, batch[1], metrics
                    )
            for k, v in training_batch_metrics.items():
                training_losses_and_metrics[k] = v

        state = self._callback_loop(state, callbacks, "training_step_end")
        return state, training_losses_and_metrics

    def training_epoch_start(
        self, state: TrainState, callbacks: Optional[List[Callback]] = None
    ) -> TrainState:
        return self._callback_loop(state, callbacks, "training_epoch_start")

    def training_epoch_end(
        self,
        training_losses_and_metrics_current_epoch: List[Dict[str, jnp.ndarray]],
        state: TrainState,
        callbacks: Optional[List[Callback]] = None,
    ) -> Tuple[TrainState, Dict[str, float]]:
        mean_losses_and_metrics = self._get_mean_losses_and_metrics(
            training_losses_and_metrics_current_epoch
        )
        state = self._callback_loop(state, callbacks, "training_epoch_end")
        return state, mean_losses_and_metrics

    def validation_epoch_end(
        self,
        validation_losses_and_metrics_current_epoch: List[Dict[str, jnp.ndarray]],
        state: TrainState,
    ) -> Dict[str, float]:
        validation_losses_and_metrics_current_epoch = self._get_mean_losses_and_metrics(
            validation_losses_and_metrics_current_epoch
        )
        # early stopping
        improved = self.early_stopping_update(
            validation_losses_and_metrics_current_epoch
        )
        if improved and self.save_checkpoint_dir:
            self.save_checkpoint(state, self.save_checkpoint_dir, force_save=True)
        return validation_losses_and_metrics_current_epoch

    def train(
        self,
        rng: PRNGKeyArray,
        state: TrainState,
        loss_fun: Callable,
        training_dataloader: DataLoader,
        training_dataset_size: int,
        n_epochs: int = 1,
        metrics: Optional[Tuple[Callable[[jnp.ndarray, Array], float], ...]] = None,
        validation_dataloader: Optional[DataLoader] = None,
        validation_dataset_size: Optional[int] = None,
        verbose: bool = True,
        unravel: Optional[Callable[[any], PyTree]] = None,
        callbacks: Optional[List[Callback]] = None,
        **kwargs,
    ) -> Tuple[TrainState, Status]:
        training_kwargs = FrozenDict(kwargs)
        if validation_dataloader:
            assert (
                validation_dataset_size is not None
            ), "`validation_dataset_size` is required when `validation_dataloader` is provided."

        training_losses_and_metrics = collections.defaultdict(list)
        validation_losses_and_metrics = collections.defaultdict(list)

        state, dataloaders, rng = self.on_train_start(
            state, [training_dataloader, validation_dataloader], rng
        )
        training_dataloader, validation_dataloader = dataloaders

        progress_bar = trange(n_epochs, desc="Epoch")
        for epoch in progress_bar:
            # training loop
            (
                state,
                training_losses_and_metrics_current_epoch,
                training_batch_metrics_str,
            ) = self._training_loop(
                epoch,
                loss_fun,
                metrics,
                rng,
                state,
                training_dataloader,
                training_dataset_size,
                training_kwargs,
                verbose,
                progress_bar,
                unravel=unravel,
                callbacks=callbacks,
            )
            # keep track of training losses and metrics [granularity=epoch]
            for k in training_losses_and_metrics_current_epoch.keys():
                training_losses_and_metrics[k].append(
                    training_losses_and_metrics_current_epoch[k]
                )

            # validation loop
            if self.should_perform_validation(validation_dataloader, epoch):
                # performance evaluation on the whole validation dataset
                state = self.on_validation_start(state)
                (
                    validation_losses_and_metrics_current_epoch,
                    validation_epoch_metrics_str,
                ) = self._validation_loop(
                    loss_fun=loss_fun,
                    metrics=metrics,
                    rng=rng,
                    state=state,
                    training_kwargs=training_kwargs,
                    validation_dataloader=validation_dataloader,
                    validation_dataset_size=validation_dataset_size,
                    verbose=verbose,
                    unravel=unravel,
                )
                if verbose:
                    logging.info(
                        f"Epoch: {epoch + 1} | " + validation_epoch_metrics_str
                    )
                # keep track of training losses and metrics [granularity=epoch] and check for early stopping
                for k in validation_losses_and_metrics_current_epoch.keys():
                    validation_losses_and_metrics[k].append(
                        validation_losses_and_metrics_current_epoch[k]
                    )
                # check for early stopping
                if self.is_early_stopping_active and self._early_stopping.should_stop:
                    logging.info("[Early Stopping] Stopping training...")
                    break

        # aggregate
        training_status = {
            k: jnp.array(v) for k, v in training_losses_and_metrics.items()
        }
        validation_status = {
            k: jnp.array(v) for k, v in validation_losses_and_metrics.items()
        }
        status = dict(**training_status, **validation_status)

        state = self.on_train_end(state)
        return state, status

    def _training_loop(
        self,
        current_epoch: int,
        loss_fun: Callable,
        metrics: Optional[Tuple[Callable[[jnp.ndarray, Array], jnp.ndarray], ...]],
        rng: PRNGKeyArray,
        state: TrainState,
        training_dataloader: DataLoader,
        training_dataset_size: int,
        training_kwargs: FrozenDict[str, Any],
        verbose: bool,
        progress_bar: TqdmDecorator,
        unravel: Optional[Callable[[any], PyTree]] = None,
        callbacks: Optional[List[Callback]] = None,
    ) -> Tuple[TrainState, Dict[str, float], str]:
        gradient_accumulation_steps = (
            training_kwargs.get("gradient_accumulation_steps", 0) or 0
        )
        training_losses_and_metrics_epoch_all_steps = []
        training_batch_metrics_str = ""
        state = self.training_epoch_start(state, callbacks)
        # ensure to use a different key at each step
        model_key = self.training_step_start(rng, state.step)
        for step, batch in enumerate(training_dataloader):
            # forward and backward pass
            state, aux = self.training_step(
                state,
                batch,
                loss_fun,
                model_key,
                training_dataset_size,
                unravel,
                training_kwargs,
            )
            self._global_training_step += 1
            if (gradient_accumulation_steps > 0) and (
                (step + 1) % gradient_accumulation_steps != 0
            ):
                continue
            # update model key
            model_key = self.training_step_start(rng, state.step)
            # compute training losses and metrics for the current batch
            state, training_losses_and_metrics_current_batch = self.training_step_end(
                current_epoch=current_epoch,
                state=state,
                aux=aux,
                batch=batch,
                metrics=metrics,
                callbacks=callbacks,
                kwargs=training_kwargs,
            )
            # keep track of training losses and metrics [granularity=batch]
            training_losses_and_metrics_epoch_all_steps.append(
                training_losses_and_metrics_current_batch
            )
            # logging
            if verbose:
                training_batch_metrics_str = " | ".join(
                    [
                        f"{m}: {round(float(v), 5)}"
                        for m, v in training_losses_and_metrics_current_batch.items()
                    ]
                )
                progress_bar.set_description(
                    f"Epoch: {current_epoch + 1} | " + training_batch_metrics_str,
                    refresh=True,
                )

        # compute training losses and metrics avg for the current epoch + other ops (if needed)
        state, training_losses_and_metrics_current_epoch = self.training_epoch_end(
            training_losses_and_metrics_epoch_all_steps, state, callbacks
        )

        return (
            state,
            training_losses_and_metrics_current_epoch,
            training_batch_metrics_str,
        )

    def _validation_loop(
        self,
        loss_fun: Callable,
        metrics: Optional[Tuple[Callable[[jnp.ndarray, Array], float], ...]],
        rng: PRNGKeyArray,
        state: TrainState,
        training_kwargs: FrozenDict[str, Any],
        validation_dataloader: DataLoader,
        validation_dataset_size: int,
        verbose: bool = True,
        unravel: Optional[Callable[[any], PyTree]] = None,
    ) -> Tuple[Dict[str, float], str]:
        validation_losses_and_metrics_epoch_all_steps = []
        validation_epoch_metrics_str = ""
        for batch in validation_dataloader:
            validation_losses_and_metrics_current_batch = self.validation_step(
                state,
                batch,
                loss_fun,
                rng,
                validation_dataset_size,
                metrics,
                unravel,
                training_kwargs,
            )
            validation_losses_and_metrics_epoch_all_steps.append(
                validation_losses_and_metrics_current_batch
            )
        # compute validation losses and metrics for the current epoch
        validation_losses_and_metrics_current_epoch = self.validation_epoch_end(
            validation_losses_and_metrics_epoch_all_steps, state
        )
        # logging
        if verbose:
            validation_epoch_metrics_str = " | ".join(
                [
                    f"{m}: {round(float(v), 5)}"
                    for m, v in validation_losses_and_metrics_current_epoch.items()
                ]
            )
        return validation_losses_and_metrics_current_epoch, validation_epoch_metrics_str

    def _callback_loop(
        self, state: TrainState, callbacks: Optional[List[Callback]], method_name: str
    ) -> TrainState:
        callbacks = callbacks or []
        for callback in callbacks:
            state = getattr(callback, method_name)(state)
        return state

    def _get_mean_losses_and_metrics(
        self, losses_and_metrics: List[Dict[str, jnp.ndarray]]
    ) -> Dict[str, float]:
        losses_and_metrics = stack_forest(losses_and_metrics)
        losses_and_metrics = tree_map(lambda x: x.mean(), losses_and_metrics)
        return losses_and_metrics

    def should_perform_validation(
        self, validation_dataloader: Optional[DataLoader], epoch: int
    ) -> bool:
        return (
            validation_dataloader is not None
            and self.eval_every_n_epochs > 0
            and epoch % self.eval_every_n_epochs == 0
        )

    @staticmethod
    def _sync_array(arr: jnp.ndarray) -> jnp.ndarray:
        return arr

    def on_train_start(
        self,
        state: TrainState,
        dataloaders: List[DataLoader],
        rng: PRNGKeyArray,
    ) -> Tuple[TrainState, List[DataLoader], PRNGKeyArray]:
        if self.freeze_fun is not None:
            frozen_paths = get_frozen_paths(state.params, self.freeze_fun)
            trainable_paths = get_trainable_paths(state.params, self.freeze_fun)
            state = state.replace(
                frozen_params=FrozenDict(
                    nested_set(
                        d={},
                        key_paths=frozen_paths,
                        objs=tuple(
                            [
                                nested_get(state.params.unfreeze(), path)
                                for path in frozen_paths
                            ]
                        ),
                        allow_nonexistent=True,
                    )
                ),
                params=FrozenDict(
                    nested_set(
                        d={},
                        key_paths=trainable_paths,
                        objs=tuple(
                            [
                                nested_get(state.params.unfreeze(), path)
                                for path in trainable_paths
                            ]
                        ),
                        allow_nonexistent=True,
                    )
                ),
            )
        return state, dataloaders, rng

    def on_train_end(self, state: TrainState) -> TrainState:
        self.save_checkpoint(
            state,
            save_checkpoint_dir=self.save_checkpoint_dir,
            keep=self.keep_top_n_checkpoints,
            force_save=True,
        )

        if self.freeze_fun is not None:
            state = state.replace(
                params=self._get_all_params(state), frozen_params=None
            )
        return state

    def on_validation_start(self, state: TrainState) -> TrainState:
        return state

    def compute_metrics(
        self,
        preds: Array,
        targets: Array,
        metrics: Optional[Tuple[Callable[[jnp.ndarray, Array], float], ...]],
        uncertainties: Optional[Array] = None,
    ) -> Dict[str, float]:
        metrics_vals = {}
        for metric in metrics:
            metrics_vals[metric.__name__] = (
                metric(preds, targets)
                if uncertainties is None
                else metric(preds, uncertainties, targets)
            )
        return metrics_vals

    def training_step_start(
        self, rng: PRNGKeyArray, step: Union[int, jax.Array]
    ) -> PRNGKeyArray:
        step = step if isinstance(step, int) or step.ndim == 0 else step[0]
        return random.fold_in(rng, step)

    def _sync_state(self, state: TrainState) -> TrainState:
        return state

    def save_checkpoint(
        self,
        state: TrainState,
        save_checkpoint_dir: Path,
        keep: int = 1,
        force_save: bool = False,
        prefix: str = "checkpoint_",
    ) -> None:
        if self.freeze_fun is not None:
            state = state.replace(
                params=self._get_all_params(state), frozen_params=None
            )
        return super().save_checkpoint(
            self._sync_state(state), save_checkpoint_dir, keep, force_save, prefix
        )

    def _get_all_params(
        self, state: TrainState, trainable_params: Optional[Params] = None
    ):
        if self.freeze_fun is not None:
            return FrozenDict(
                nested_update(
                    state.frozen_params.unfreeze(),
                    trainable_params.unfreeze()
                    if trainable_params is not None
                    else state.params.unfreeze(),
                )
            )
        return trainable_params if trainable_params is not None else state.params


class JittedMixin:
    @partial(jax.jit, static_argnums=(0, 3, 5, 6, 7))
    def training_step(
        self,
        state: TrainState,
        batch: Batch,
        loss_fun: Callable,
        rng: PRNGKeyArray,
        n_data: int,
        unravel: Optional[Callable[[any], PyTree]] = None,
        kwargs: FrozenDict[str, Any] = FrozenDict(),
    ) -> Tuple[TrainState, Dict[str, Any]]:
        return super().training_step(
            state, batch, loss_fun, rng, n_data, unravel, kwargs
        )

    @partial(jax.jit, static_argnums=(0, 3, 5, 6, 7, 8))
    def validation_step(
        self,
        state: TrainState,
        batch: Batch,
        loss_fun: Callable,
        rng: PRNGKeyArray,
        n_data: int,
        metrics: Optional[Tuple[Callable[[jnp.ndarray, Array], float], ...]] = None,
        unravel: Optional[Callable[[any], PyTree]] = None,
        kwargs: FrozenDict[str, Any] = FrozenDict(),
    ) -> Dict[str, jnp.ndarray]:
        return super().validation_step(
            state, batch, loss_fun, rng, n_data, metrics, unravel, kwargs
        )


class MultiDeviceMixin:
    all_reduce_mean = jax.pmap(lambda x: lax.pmean(x, "x"), "x")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.multi_device = True

    @staticmethod
    def _add_device_dim_to_input_dataloader(dataloader: DataLoader) -> DataLoader:
        def _reshape_input_batch(batch):
            n_devices = jax.local_device_count()
            if batch.shape[0] % n_devices != 0:
                raise ValueError(
                    f"The size of all batches must be a multiple of {n_devices}, that is the number of "
                    f"available devices. Please set an appropriate batch size in the data loader."
                )
            single_input_shape = batch.shape[1:]
            # reshape to (local_devices, device_batch_size, *single_input_shape)
            return batch.reshape((n_devices, -1) + single_input_shape)

        class DataLoaderWrapper:
            def __init__(self, dataloader):
                self.dataloader = dataloader

            def __iter__(self):
                dataloader = map(
                    lambda batch: tree_map(_reshape_input_batch, batch), self.dataloader
                )
                dataloader = jax_utils.prefetch_to_device(dataloader, 2)
                yield from dataloader

        return DataLoaderWrapper(dataloader) if dataloader is not None else dataloader

    @staticmethod
    def _sync_mutable(state: TrainState) -> TrainState:
        return (
            state.replace(mutable=MultiDeviceMixin.all_reduce_mean(state.mutable))
            if state.mutable is not None
            else state
        )

    @staticmethod
    def _sync_array(arr: jnp.ndarray) -> jnp.ndarray:
        arr = lax.pmean(arr, axis_name="batch")
        return arr

    def _sync_state(self, state: TrainState) -> TrainState:
        state = self._sync_mutable(state)
        return jax.device_get(tree_map(lambda x: x[0], state))

    def on_train_start(
        self, state: TrainState, dataloaders: List[DataLoader], rng: PRNGKeyArray
    ) -> Tuple[TrainState, List[DataLoader], PRNGKeyArray]:
        state, dataloaders, rng = super(MultiDeviceMixin, self).on_train_start(
            state, dataloaders, rng
        )
        state = jax_utils.replicate(state)
        dataloaders = [
            self._add_device_dim_to_input_dataloader(dl) for dl in dataloaders
        ]
        model_key = random.split(rng, jax.local_device_count())
        return state, dataloaders, model_key

    def on_train_end(self, state: TrainState) -> TrainState:
        state = super(MultiDeviceMixin, self).on_train_end(state)
        return jax.device_get(tree_map(lambda x: x[0], state))

    def training_step_start(self, rng: PRNGKeyArray, step: int) -> PRNGKeyArray:
        step = step if isinstance(step, int) or step.ndim == 0 else step[0]
        return jax.vmap(lambda r: random.fold_in(r, step))(rng)

    @partial(jax.pmap, axis_name="batch", static_broadcasted_argnums=(0, 3, 5, 6, 7))
    def training_step(
        self,
        state: TrainState,
        batch: Batch,
        loss_fun: Callable,
        rng: PRNGKeyArray,
        n_data: int,
        unravel: Optional[Callable[[any], PyTree]] = None,
        kwargs: FrozenDict[str, Any] = FrozenDict(),
    ) -> Tuple[TrainState, Dict[str, Any]]:
        return super().training_step(
            state, batch, loss_fun, rng, n_data, unravel, kwargs
        )

    def training_step_end(
        self,
        current_epoch: int,
        state: TrainState,
        aux: Dict[str, Any],
        batch: Batch,
        metrics: Optional[Tuple[Callable[[jnp.ndarray, Array], float], ...]],
        callbacks: Optional[List[Callback]] = None,
        kwargs: FrozenDict[str, Any] = FrozenDict(),
    ) -> Tuple[TrainState, Dict[str, jnp.ndarray]]:
        state, training_losses_and_metrics = super(
            MultiDeviceMixin, self
        ).training_step_end(
            current_epoch, state, aux, batch, metrics, callbacks, kwargs
        )
        return state, tree_map(lambda x: x.mean(), training_losses_and_metrics)

    def on_validation_start(self, state: TrainState) -> TrainState:
        state = super(MultiDeviceMixin, self).on_validation_start(state)
        state = self._sync_mutable(state)
        return state

    @partial(jax.pmap, axis_name="batch", static_broadcasted_argnums=(0, 3, 5, 6, 7, 8))
    def validation_step(
        self,
        state: TrainState,
        batch: Batch,
        loss_fun: Callable,
        rng: PRNGKeyArray,
        n_data: int,
        metrics: Optional[Tuple[Callable[[jnp.ndarray, Array], float], ...]] = None,
        unravel: Optional[Callable[[any], PyTree]] = None,
        kwargs: FrozenDict[str, Any] = FrozenDict(),
    ) -> Dict[str, jnp.ndarray]:
        validation_losses_and_metrics = super().validation_step(
            state, batch, loss_fun, rng, n_data, metrics, unravel, kwargs
        )
        return lax.pmean(validation_losses_and_metrics, axis_name="batch")



"""
Adapted from https://github.com/omegafragger/DDU/blob/main/utils/gmm_utils.py
"""
import logging
from typing import (
    Callable,
    Tuple,
)

import jax
from jax import numpy as jnp
import jax.scipy as jsp
import jax.scipy.stats as jsp_stats
import numpy as np

from fortuna.ood_detection.base import (
    NotFittedError,
    OutOfDistributionClassifierABC,
)
from fortuna.typing import Array

DOUBLE_INFO = np.finfo(np.double)
JITTERS = [0, DOUBLE_INFO.tiny] + [10**exp for exp in range(-10, 0, 1)]


def _centered_cov(x: Array) -> Array:
    n = x.shape[0]
    res = jnp.matmul(1 / (n - 1) * x.T, x)
    return res


def compute_classwise_mean_and_cov(
    embeddings: Array, labels: Array, num_classes: int
) -> Tuple[Array, Array]:
    """
    Computes class-specific means and a covariance matrices given the training set embeddings
    (e.g., the last-layer representation of the model for each training example).

    Parameters
    ----------
    embeddings: Array
        The embeddings of shape `(n, d)` where `n` is the number of training samples and `d` is the embbeding's size.
    labels: Array
        An array of length `n` containing, for each input sample, its ground-truth label.
    num_classes: int
        The total number of classes available in the classification task.

    Returns
    ----------
    Tuple[Array, Array]:
        A tuple containing:
        1) an `Array` containing the per-class mean vector of the fitted GMM.
        The shape of the array is `(num_classes, d)`.
        2) an `Array` containing the per-class covariance matrix of the fitted GMM.
        The shape of the array is `(num_classes, d, d)`.
    """
    #
    classwise_mean_features = np.stack(
        [jnp.mean(embeddings[labels == c], 0) for c in range(num_classes)]
    )
    #
    classwise_cov_features = np.stack(
        [
            _centered_cov(embeddings[labels == c] - classwise_mean_features[c])
            for c in range(num_classes)
        ]
    )
    return classwise_mean_features, classwise_cov_features


def _get_logpdf_fn(
    classwise_mean_features: Array, classwise_cov_features: Array
) -> Callable[[Array], Array]:
    """
    Returns a function to evaluate the log-likelihood of a test sample according to the (fitted) GMM.

    Parameters
    ----------
    classwise_mean_features: Array
        The per-class mean vector of the fitted GMM. The shape of the array is `(num_classes, d)`.
    classwise_cov_features: Array
        The per-class covariance matrix of the fitted GMM. The shape of the array is `(num_classes, d, d)`.

    Returns
    -------
    Callable[[Array], Array]
        A function to evaluate the log-likelihood of a test sample according to the (fitted) GMM.
    """
    for jitter_eps in JITTERS:
        jitter = np.expand_dims(jitter_eps * np.eye(classwise_cov_features.shape[1]), 0)
        gmm_logprob_fn_vmapped = jax.vmap(
            jsp_stats.multivariate_normal.logpdf, in_axes=(None, 0, 0)
        )
        gmm_logprob_fn = lambda x: gmm_logprob_fn_vmapped(
            x, classwise_mean_features, (classwise_cov_features + jitter)
        ).T

        nans = np.isnan(gmm_logprob_fn(classwise_mean_features)).sum()
        if nans > 0:
            logging.info(f"Nans, jittering {jitter_eps}")
            continue
        break

    return gmm_logprob_fn


class DeepDeterministicUncertaintyOODClassifier(OutOfDistributionClassifierABC):
    """
    A Gaussian Mixture Model :math:`q(\mathbf{x}, z)` with a single Gaussian mixture component per class :math:`k \in {1,...,K}`
    is fit after training.
    Each class component is fit computing the empirical mean :math:`\mathbf{\hat{\mu}_k}` and covariance matrix
    :math:`\mathbf{\hat{\Sigma}_k}` of the feature vectors :math:`f(\mathbf{x})`.

    The confidence score :math:`M(\mathbf{x})` for a new test sample is obtained computing the negative marginal likelihood
    of the feature representation.

    See `Mukhoti, Jishnu, et al. <https://arxiv.org/abs/2102.11582>`_
    """

    def __init__(self, *args, **kwargs):
        super(DeepDeterministicUncertaintyOODClassifier, self).__init__(*args, **kwargs)
        self._classwise_mean_features = None
        self._classwise_cov_features = None
        self._gmm_logpdf_fn = None

    @property
    def mean(self) -> Array:
        """
        Returns
        -------
        Array
            The per-class mean vector of the fitted GMM. The shape of the array is `(num_classes, d)`
            where `num_classes` is the number of target classes in the in-distribution classification task
            and `d` is the embedding size.
        """
        return self._classwise_mean_features

    @property
    def cov(self):
        """
        Returns
        -------
        Array
            The per-class covariance matrix of the fitted GMM. The shape of the array is `(num_classes, d, d)`
            where `num_classes` is the number of target classes in the in-distribution classification task
            and `d` is the embedding size.
        """
        return self._classwise_cov_features

    def fit(self, embeddings: Array, targets: Array) -> None:
        """
        Fits a Multivariate Gaussian to the training data using class-specific means and covariance matrix.

        Parameters
        ----------
        embeddings: Array
            The embeddings of shape `(n, d)` where `n` is the number of training samples and `d` is the embbeding's size.
        targets: Array
            An array of length `n` containing, for each input sample, its ground-truth label.
        """
        (
            self._classwise_mean_features,
            self._classwise_cov_features,
        ) = compute_classwise_mean_and_cov(embeddings, targets, self.num_classes)
        self._gmm_logpdf_fn = _get_logpdf_fn(
            self._classwise_mean_features, self._classwise_cov_features
        )

    def score(self, embeddings: Array) -> Array:
        """
        The confidence score :math:`M(\mathbf{x})` for a new test sample :math:`\mathbf{x}` is obtained computing
        the negative marginal likelihood of the feature representation
        :math:`-q(f(\mathbf{x})) = - \sum\limits_{k}q(f(\mathbf{x})|y) q(y)`.

        A high score signals that the test sample :math:`\mathbf{x}` is identified as OOD.

        Parameters
        ----------
        embeddings: Array
            The embeddings of shape `(n, d)` where `n` is the number of test samples and `d` is the embbeding's size.

        Returns
        -------
        Array
            An array of scores with length `n`.
        """
        if self._gmm_logpdf_fn is None:
            raise NotFittedError("You have to call fit before calling score.")
        loglik = self._gmm_logpdf_fn(embeddings)
        return -jsp.special.logsumexp(jnp.nan_to_num(loglik, 0.0), axis=1)



import jax.numpy as jnp
from jax.tree_util import (
    tree_map,
    tree_reduce,
)


def clip_grandients_by_value(grad: jnp.ndarray, max_grad_val: float) -> jnp.ndarray:
    clip_fn = lambda z: jnp.clip(z, -max_grad_val, max_grad_val)
    grad = tree_map(clip_fn, grad)
    return grad


def clip_grandients_by_norm(grad: jnp.ndarray, max_grad_norm: float) -> jnp.ndarray:
    grad_norm = jnp.sqrt(
        tree_reduce(lambda x, y: x + jnp.sum(y**2), grad, initializer=0)
    )
    mult = jnp.minimum(1, max_grad_norm / (1e-7 + grad_norm))
    grad = tree_map(lambda z: mult * z, grad)
    return grad



import abc
import logging
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

from datasets import (
    Dataset,
    DatasetDict,
)
from jax import numpy as jnp
import jax.random
from jax.random import PRNGKeyArray
from tqdm import tqdm
from transformers import (
    BatchEncoding,
    DataCollatorWithPadding,
    PreTrainedTokenizer,
)
from transformers.utils import PaddingStrategy

from fortuna.data.dataset.data_collator import (
    FlaxDataCollatorForLanguageModeling,
    FlaxDataCollatorForMultipleChoice,
)
from fortuna.data.loader.huggingface_loaders import HuggingFaceDataLoader
from fortuna.data.loader.utils import IterableData
from fortuna.typing import Array

logger = logging.getLogger(__name__)


class HuggingFaceClassificationDatasetABC(abc.ABC):
    def __init__(
        self,
        tokenizer: PreTrainedTokenizer,
        max_length: int,
        padding: Union[bool, str, PaddingStrategy] = True,
        pad_to_multiple_of: int = 64,
        num_unique_labels: Optional[int] = None,
    ):
        """
        A Dataset class to work with HuggingFace-like datasets.

        Parameters
        ----------
        tokenizer: PreTrainedTokenizer
            A pretrained tokenizer
        max_length: int
            Maximum sequence length (equal for each batch).
            It should be se to max length accepted by the model or a smaller number.
        padding: Union[bool, str, PaddingStrategy]
            See `Padding and Truncation <https://huggingface.co/docs/transformers/pad_truncation>`_
            for more information (`truncation` is always True).
        pad_to_multiple_of: int
            Pad the sequence to a multiple of the provided value.
            This argument will be used whenever padding is not done to a fixed constant (e.g., max_length).
        num_unique_labels: Optional[int]
            Number of unique target labels in the task (classification only)
        """
        self.tokenizer = tokenizer
        self.max_length = max_length
        self.padding = padding
        self.pad_to_multiple_of = pad_to_multiple_of
        self.num_unique_labels = num_unique_labels
        self._data_collator = None

    @abc.abstractmethod
    def get_tokenized_datasets(
        self, datasets: DatasetDict, *args, **kwargs
    ) -> DatasetDict:
        pass

    @property
    @abc.abstractmethod
    def data_collator(self):
        """
        Create a data collator object that will process batches of data and apply transformation on them if needed.
        """
        pass

    def get_data_loader(
        self,
        dataset: Dataset,
        per_device_batch_size: int,
        rng: PRNGKeyArray,
        shuffle: bool = False,
        drop_last: bool = False,
        verbose: bool = False,
    ) -> HuggingFaceDataLoader:
        """
        Build a :class:`~fortuna.data.loader.huggingface_loaders.HuggingFaceDataLoader` object from a
        tokenized dataset.

        Parameters
        ----------
        dataset: Dataset
            A tokenizeed dataset (see :meth:`.HuggingFaceClassificationDatasetABC.get_tokenized_datasets`).
        per_device_batch_size: bool
            Batch size for each device.
        rng: PRNGKeyArray
            Random number generator.
        shuffle: bool
            if True, shuffle the data so that each batch is a ranom sample from the dataset.
        drop_last: bool
            if True, the last batch (which potentially is smaller then the default batch size) is dropped.
        verbose: bool
            Whether to show a progress bar while iterating over the dataloader or not.

        Returns
        -------
        HuggingFaceDataLoader
            The dataloader
        """
        iterable = IterableData.from_callable(
            lambda *args, **kwargs: self._get_data_loader(
                dataset,
                batch_size=per_device_batch_size * jax.local_device_count(),
                shuffle=shuffle,
                drop_last=drop_last,
                rng=rng,
                verbose=verbose,
            )
        )
        return HuggingFaceDataLoader(
            iterable=iterable,
            num_inputs=len(dataset),
            num_unique_labels=self.num_unique_labels,
        )

    @abc.abstractmethod
    def _collate(self, batch: Dict[str, Array], batch_size: int) -> Dict[str, Array]:
        pass

    @staticmethod
    def _get_batches_idxs(
        rng: PRNGKeyArray,
        dataset_size: int,
        batch_size: int,
        shuffle: bool = False,
        drop_last: bool = False,
    ) -> Iterable[Array]:
        if shuffle:
            dataset_idxs = jax.random.permutation(rng, dataset_size)  # batch idxs
        else:
            dataset_idxs = jnp.arange(dataset_size)
        if drop_last:
            steps_per_epoch = dataset_size // batch_size
            dataset_idxs = dataset_idxs[
                : steps_per_epoch * batch_size
            ]  # Skip incomplete batch.
        yield from [
            dataset_idxs[start : start + batch_size]
            for start in range(0, len(dataset_idxs), batch_size)
        ]

    def _get_data_loader(
        self,
        dataset: Dataset,
        batch_size: int,
        shuffle: bool = False,
        drop_last: bool = False,
        rng: Optional[PRNGKeyArray] = None,
        verbose: bool = False,
    ) -> Union[Iterable[Dict[str, Array]], Iterable[Tuple[Dict[str, Array], Array]]]:
        batch_idxs_gen = self._get_batches_idxs(
            rng=rng,
            dataset_size=len(dataset),
            batch_size=batch_size,
            shuffle=shuffle,
            drop_last=drop_last,
        )
        if verbose:
            steps_per_epoch = len(dataset) // batch_size
            steps_per_epoch = steps_per_epoch if drop_last else steps_per_epoch + 1
            batch_idxs_gen = tqdm(
                batch_idxs_gen, desc="Batches:", total=steps_per_epoch
            )
        for batch_idxs in batch_idxs_gen:
            batch = dataset[batch_idxs]
            batch = self._collate(batch, len(batch["input_ids"]))
            batch_inputs = {k: jnp.array(v) for k, v in batch.items()}
            if "labels" in batch_inputs:
                batch_labels = jnp.array(batch_inputs.pop("labels"))
                yield (batch_inputs, batch_labels)
            else:
                yield batch_inputs, None


class HuggingFaceSequenceClassificationDataset(HuggingFaceClassificationDatasetABC):
    """
    Dataset for Sequence Classification tasks.
    """

    def get_tokenized_datasets(
        self,
        datasets: DatasetDict,
        text_columns: Sequence[str] = ("sentence",),
        target_column: str = "label",
        **kwargs,
    ) -> DatasetDict:
        """
        Converts one or more dataset into sequences of tokens, using the tokenizer.

        Parameters
        ----------
        datasets: DatasetDict
            A dictionary of `datasets.Dataset` that have to be encoded.
        text_columns: str
            A list containing the text column names, whose text sequences has to be encoded.
        target_column: str
            The target column name

        Returns
        -------
        DatasetDict
            A dictionary of tokenized datasets.
        """

        def _tokenize_fn(batch: Dict[str, List[Union[str, int]]]) -> BatchEncoding:
            tokenized_inputs = self.tokenizer(
                *[batch[col] for col in text_columns],
                padding="max_length" if self.padding == "max_length" else False,
                max_length=self.max_length,
                truncation=True,
            )
            if target_column in batch:
                tokenized_inputs["label"] = batch[target_column]
            return tokenized_inputs

        tokenized_datasets = datasets.map(
            _tokenize_fn,
            batched=True,
            remove_columns=datasets[list(datasets.keys())[0]].column_names,
        )
        return tokenized_datasets

    @property
    def data_collator(self) -> Any:
        if self._data_collator is None:
            self._data_collator = DataCollatorWithPadding(
                tokenizer=self.tokenizer,
                padding=self.padding,
                max_length=self.max_length,
                return_tensors="np",
            )
        return self._data_collator

    def _collate(self, batch: Dict[str, Array], batch_size: int) -> Dict[str, Array]:
        if self.padding and self.padding != "max_length":
            batch = self.data_collator(batch)
        else:
            if "label" in batch:
                batch["labels"] = batch["label"]
                del batch["label"]
        return batch


class HuggingFaceMultiChoiceDataset(HuggingFaceClassificationDatasetABC):
    """
    Dataset for Multi Choice classification tasks.
    """

    def get_tokenized_datasets(
        self,
        datasets: DatasetDict,
        contexts: Sequence[str] = (),
        choices: Sequence[str] = (),
        target_column: str = "label",
        **kwargs,
    ) -> DatasetDict:
        """
        Converts one or more dataset into sequences of tokens, using the tokenizer.

        Parameters
        ----------
        datasets: DatasetDict
            A dictionary of `datasets.Dataset` that have to be encoded.
        contexts: Sequence[str]
            A list of str containing the column names for the contexts.
            The first column contain the initial context (the first sentence).
            The second column name contains the beginning of the second sentence.
        choices:
            A list of str containing the column names for the possible continuations of the context.
        target_column: str
            The target column name
        Returns
        -------
        DatasetDict
            A dictionary of tokenized datasets.
        """
        sent1 = contexts[0]
        sent2 = contexts[1]

        def _tokenize_fn(batch: Dict[str, List[Union[str, int]]]) -> Dict:
            first_sentences = [
                [context] * self.num_unique_labels for context in batch[sent1]
            ]
            question_headers = batch[sent2]
            if len(choices) > 1:
                second_sentences = [
                    [f"{header} {batch[end][i]}" for end in choices]
                    for i, header in enumerate(question_headers)
                ]
            else:
                second_sentences = [
                    [f"{header} {end}" for end in batch[choices[0]][i]]
                    for i, header in enumerate(question_headers)
                ]
            first_sentences = sum(first_sentences, [])
            second_sentences = sum(second_sentences, [])
            tokenized_inputs = self.tokenizer(
                first_sentences, second_sentences, truncation=True
            )
            tokenized_inputs = {
                k: [
                    v[i : i + self.num_unique_labels]
                    for i in range(0, len(v), self.num_unique_labels)
                ]
                for k, v in tokenized_inputs.items()
            }
            tokenized_inputs["label"] = batch[target_column]
            return tokenized_inputs

        tokenized_datasets = datasets.map(
            _tokenize_fn,
            batched=True,
            remove_columns=datasets[list(datasets.keys())[0]].column_names,
        )
        return tokenized_datasets

    @property
    def data_collator(self) -> Any:
        if self._data_collator is None:
            self._data_collator = FlaxDataCollatorForMultipleChoice(
                tokenizer=self.tokenizer,
                max_length=self.max_length,
                pad_to_multiple_of=self.pad_to_multiple_of,
            )
        return self._data_collator

    def _collate(self, batch: Dict[str, Array], batch_size: int) -> Dict[str, Array]:
        batch = [{k: batch[k][i] for k in batch.keys()} for i in range(batch_size)]
        batch = self.data_collator(batch)
        return batch


class HuggingFaceMaskedLMDataset(HuggingFaceClassificationDatasetABC):
    def __init__(
        self,
        *args,
        mlm: bool = True,
        mlm_probability: Optional[float] = 0.15,
        **kwargs,
    ):
        """
        Parameters
        ----------
        mlm: bool
            Whether or not to use masked language modeling. If set to `False`, the labels are the same as the inputs
            with the padding tokens ignored (by setting them to -100). Otherwise, the labels are -100 for non-masked
            tokens and the value to predict for the masked token. Default: True.
        mlm_probability Optional[float]
            The probability with which to (randomly) mask tokens in the input,
            when the task is masked language modeling.
        """
        _ = kwargs.pop("num_unique_labels", None)
        super(HuggingFaceMaskedLMDataset, self).__init__(*args, **kwargs)
        if not self.tokenizer.is_fast:
            logger.warning(
                f"You are not using a Fast Tokenizer, so whole words cannot be masked, only tokens."
            )
        self.mlm = mlm
        self.mlm_probability = mlm_probability
        self.num_unique_labels = self.tokenizer.vocab_size

    def get_tokenized_datasets(
        self,
        datasets: DatasetDict,
        text_columns: Sequence[str] = ("sentence",),
        **kwargs,
    ) -> DatasetDict:
        """
        Converts one or more dataset into sequences of tokens, using the tokenizer.

        Parameters
        ----------
        datasets: DatasetDict
            A dictionary of `datasets.Dataset` that have to be encoded.
        text_columns: str
            A list containing the text column names, whose text sequences has to be encoded.

        Returns
        -------
        DatasetDict
            A dictionary of tokenized datasets.
        """
        assert (
            len(text_columns) == 1
        ), "Only one text column should be passed when the task is MaskedLM."

        def _tokenize_fn(
            batch: Dict[str, List[Union[str, int]]]
        ) -> Dict[str, List[int]]:
            tokenized_inputs = self.tokenizer(
                *[batch[col] for col in text_columns],
                truncation=False,
                return_special_tokens_mask=True,
            )
            # Concatenate all texts
            keys = list(tokenized_inputs.keys())
            tokenized_inputs_cat = {k: sum(tokenized_inputs[k], []) for k in keys}
            # Compute length of concatenated texts
            total_length = len(tokenized_inputs_cat[keys[0]])
            # We drop the last chunk if it's smaller than chunk_size
            total_length = (total_length // self.max_length) * self.max_length
            # Split by chunks of max_len
            tokenized_inputs = {
                k: [
                    t[i : i + self.max_length]
                    for i in range(0, total_length, self.max_length)
                ]
                for k, t in tokenized_inputs_cat.items()
            }
            return tokenized_inputs

        tokenized_datasets = datasets.map(
            _tokenize_fn,
            batched=True,
            remove_columns=datasets[list(datasets.keys())[0]].column_names,
        )
        return tokenized_datasets

    @property
    def data_collator(self):
        if self._data_collator is None:
            self._data_collator = FlaxDataCollatorForLanguageModeling(
                tokenizer=self.tokenizer,
                mlm=self.mlm,
                mlm_probability=self.mlm_probability,
            )
        return self._data_collator

    def _collate(self, batch: Dict[str, Array], batch_size: int) -> Dict[str, Array]:
        batch = [{k: batch[k][i] for k in batch.keys()} for i in range(batch_size)]
        return self.data_collator(batch, pad_to_multiple_of=self.pad_to_multiple_of)



import abc
from typing import (
    List,
    Tuple,
)

from jax import vmap
import jax.numpy as jnp
import numpy as np

from fortuna.typing import Array


class ConformalClassifier:
    """
    A base conformal classifier class.
    """

    def is_in(self, values: Array, conformal_sets: List) -> Array:
        """
        Check whether the values lie within their respective conformal sets.

        Parameters
        ----------
        values: Array
            Values to check if they lie in the respective conformal sets.
        conformal_sets: Array
            A conformal set for each input data point.

        Returns
        -------
        Array
            An array of ones or zero, indicating whether the values lie within their respective conformal sets.
        """
        return jnp.array([v in s for v, s in zip(values.tolist(), conformal_sets)])

    @abc.abstractmethod
    def score_fn(
        self,
        probs: Array,
        targets: Array,
    ):
        pass

    @staticmethod
    def _get_conformal_sets_from_scores(
        val_scores: Array,
        test_scores: Array,
        error: float,
    ) -> List[List[int]]:
        conds = jnp.sum(val_scores[:, None, None] > test_scores[None], axis=0) < (
            1 - error
        ) * (len(val_scores) + 1)
        sizes = conds.sum(1)

        sets = np.zeros(len(test_scores), dtype=object)
        for us in jnp.unique(sizes):
            idx = jnp.where(sizes == us)[0]
            if us == 0:
                sets[idx] = [len(idx) * []]
            else:
                sets[idx] = np.where(conds[idx])[1].reshape(-1, us).tolist()

        return sets.tolist()

    @abc.abstractmethod
    def get_scores(self, *args, **kwargs) -> Tuple[Array, Array]:
        pass


class SplitConformalClassifier(ConformalClassifier, abc.ABC):
    def get_scores(
        self, val_probs: Array, val_targets: Array, test_probs: Array
    ) -> Tuple[Array, Array]:
        val_scores = self.score_fn(val_probs, val_targets)
        test_scores = vmap(
            lambda i: self.score_fn(
                test_probs, i * jnp.ones(len(test_probs), dtype="int32")
            ),
            out_axes=1,
        )(jnp.arange(val_probs.shape[1]))
        return val_scores, test_scores

    def conformal_set(
        self, val_probs: Array, val_targets: Array, test_probs: Array, error: float
    ) -> List[List[int]]:
        val_scores, test_scores = self.get_scores(
            val_probs=val_probs, val_targets=val_targets, test_probs=test_probs
        )
        return super()._get_conformal_sets_from_scores(
            val_scores=val_scores, test_scores=test_scores, error=error
        )


class CVPlusConformalClassifier(ConformalClassifier):
    def conformal_set(
        self,
        cross_val_probs: List[Array],
        cross_val_targets: List[Array],
        cross_test_probs: List[Array],
        error: float,
    ) -> List[List[int]]:
        val_scores, test_scores = self.get_scores(
            cross_val_probs=cross_val_probs,
            cross_val_targets=cross_val_targets,
            cross_test_probs=cross_test_probs,
        )
        return super()._get_conformal_sets_from_scores(
            val_scores=val_scores, test_scores=test_scores, error=error
        )

    def get_scores(
        self,
        cross_val_probs: List[Array],
        cross_val_targets: List[Array],
        cross_test_probs: List[Array],
    ) -> Tuple[Array, Array]:
        val_scores, test_scores = [], []
        for val_probs, val_targets, test_probs in zip(
            cross_val_probs, cross_val_targets, cross_test_probs
        ):
            val_scores.append(self.score_fn(val_probs, val_targets))
            test_scores.append(
                vmap(
                    lambda i: self.score_fn(
                        test_probs, i * jnp.ones(len(test_probs), dtype="int32")
                    ),
                    out_axes=1,
                )(jnp.arange(cross_val_probs[0].shape[1]))
            )

        return jnp.concatenate(val_scores), jnp.concatenate(test_scores, axis=0)



from typing import Optional

import jax
from jax import vmap
import jax.numpy as jnp

from fortuna.data.loader import InputsLoader
from fortuna.likelihood.base import Likelihood
from fortuna.model.model_manager.classification import ClassificationModelManager
from fortuna.output_calibrator.output_calib_manager.base import OutputCalibManager
from fortuna.prob_output_layer.classification import ClassificationProbOutputLayer
from fortuna.typing import (
    Array,
    CalibMutable,
    CalibParams,
    Mutable,
    Params,
)


class ClassificationLikelihood(Likelihood):
    def __init__(
        self,
        model_manager: ClassificationModelManager,
        prob_output_layer: ClassificationProbOutputLayer,
        output_calib_manager: Optional[OutputCalibManager] = None,
    ):
        """
        A classification likelihood function class. In this class, the likelihood function is additionally assumed to
        be a probability density function, i.e. positive and integrating to 1. The likelihood is formed by three
        objects applied in sequence: the model manager, the output calibrator and the probabilistic output layer. The
        model manager maps parameters and inputs to outputs. The output calibration takes outputs and returns some
        calibrated version of them. The probabilistic output layer describes the probability distribution of the
        calibrated outputs.

        Parameters
        ----------
        model_manager : ModelManager
            An model manager. This objects orchestrates the evaluation of the models.
        prob_output_layer : ProbOutputLayer
            A probabilistic output layer object. This object characterizes the probability distribution of the
            target variable given the calibrated outputs.
        output_calib_manager : Optional[OutputCalibManager]
            An output calibration manager object. It transforms outputs of the model manager into some
            calibrated version of them.
        """
        super().__init__(
            model_manager, prob_output_layer, output_calib_manager=output_calib_manager
        )

    def mean(
        self,
        params: Params,
        inputs_loader: InputsLoader,
        mutable: Optional[Mutable] = None,
        calib_params: Optional[CalibParams] = None,
        calib_mutable: Optional[CalibMutable] = None,
        distribute: bool = True,
        **kwargs,
    ) -> jnp.ndarray:
        r"""
        Estimate the likelihood mean of the one-hot encoded target variable, that is

        .. math::
            \mathbb{E}_{\tilde{Y}|w, x}[\tilde{Y}],

        where:
         - :math:`x` is an observed input variable;
         - :math:`\tilde{Y}` is a one-hot encoded random target variable;
         - :math:`w` denotes the observed model parameters.

        Parameters
        ----------
        params : Params
            The random parameters of the probabilistic model.
        inputs_loader : InputsLoader
            A loader of input data points.
        mutable : Optional[Mutable]
            The mutable objects used to evaluate the models.
        calib_params : Optional[CalibParams]
            The calibration parameters of the probabilistic model.
        calib_mutable : Optional[CalibMutable]
            The calibration mutable objects used to evaluate the calibrators.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            An estimate of the likelihood mean for each input.
        """
        return super().mean(
            params,
            inputs_loader,
            mutable,
            calib_params,
            calib_mutable,
            distribute,
            **kwargs,
        )

    def _batched_mean(
        self,
        params: Params,
        inputs: Array,
        mutable: Optional[Mutable] = None,
        calib_params: Optional[CalibParams] = None,
        calib_mutable: Optional[CalibMutable] = None,
        **kwargs,
    ) -> jnp.ndarray:
        outputs = self._get_batched_calibrated_outputs(
            params, inputs, mutable, calib_params, calib_mutable, **kwargs
        )
        return jax.nn.softmax(outputs, -1)

    def _batched_mode(
        self,
        params: Params,
        inputs: Array,
        mutable: Optional[Mutable] = None,
        calib_params: Optional[CalibParams] = None,
        calib_mutable: Optional[CalibMutable] = None,
        **kwargs,
    ) -> jnp.ndarray:
        outputs = self._get_batched_calibrated_outputs(
            params, inputs, mutable, calib_params, calib_mutable, **kwargs
        )
        return jnp.argmax(outputs, -1)

    def variance(
        self,
        params: Params,
        inputs_loader: InputsLoader,
        mutable: Optional[Mutable] = None,
        calib_params: Optional[CalibParams] = None,
        calib_mutable: Optional[CalibMutable] = None,
        distribute: bool = True,
        **kwargs,
    ) -> jnp.ndarray:
        r"""
        Estimate the likelihood variance of the one-hot encoded target variable, that is

        .. math::
            \text{Var}_{\tilde{Y}|w,x}[\tilde{Y}],

        where:
         - :math:`x` is an observed input variable;
         - :math:`\tilde{Y}` is a one-hot encoded random target variable;
         - :math:`w` denotes the observed model parameters.

        Parameters
        ----------
        params : Params
            The random parameters of the probabilistic model.
        inputs_loader : InputsLoader
            A loader of input data points.
        mutable : Optional[Mutable]
            The mutable objects used to evaluate the models.
        calib_params : Optional[CalibParams]
            The calibration parameters of the probabilistic model.
        calib_mutable : Optional[CalibMutable]
            The calibration mutable objects used to evaluate the calibrators.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            An estimate of the likelihood variance for each input.
        """
        return super().variance(
            params,
            inputs_loader,
            mutable,
            calib_params,
            calib_mutable,
            distribute,
            **kwargs,
        )

    def _batched_variance(
        self,
        params: Params,
        inputs: Array,
        mutable: Optional[Mutable] = None,
        calib_params: Optional[CalibParams] = None,
        calib_mutable: Optional[CalibMutable] = None,
        distribute: bool = True,
        **kwargs,
    ) -> jnp.ndarray:
        means = self._batched_mean(
            params=params,
            inputs=inputs,
            mutable=mutable,
            calib_params=calib_params,
            calib_mutable=calib_mutable,
            **kwargs,
        )
        return means * (1 - means)

    def entropy(
        self,
        params: Params,
        inputs_loader: InputsLoader,
        mutable: Optional[Mutable] = None,
        calib_params: Optional[CalibParams] = None,
        calib_mutable: Optional[CalibMutable] = None,
        distribute: bool = True,
        **kwargs,
    ) -> jnp.ndarray:
        outputs = super().get_calibrated_outputs(
            params, inputs_loader, mutable, calib_params, calib_mutable, distribute
        )
        n_classes = outputs.shape[-1]

        @vmap
        def _entropy_term(i: int):
            targets = i * jnp.ones(outputs.shape[0])
            log_liks = self.prob_output_layer.log_prob(outputs, targets, **kwargs)
            return jnp.exp(log_liks) * log_liks

        return -jnp.sum(_entropy_term(jnp.arange(n_classes)), 0)



from __future__ import annotations

import logging
import os
import pathlib
from typing import (
    List,
    Optional,
    Tuple,
    Type,
)

from flax.core import FrozenDict
from jax import (
    pure_callback,
    random,
)
from jax._src.prng import PRNGKeyArray

from fortuna.data.loader import DataLoader
from fortuna.prob_model.fit_config.base import FitConfig
from fortuna.prob_model.joint.base import Joint
from fortuna.prob_model.joint.state import JointState
from fortuna.prob_model.posterior.base import Posterior
from fortuna.prob_model.posterior.deep_ensemble import DEEP_ENSEMBLE_NAME
from fortuna.prob_model.posterior.deep_ensemble.deep_ensemble_approximator import (
    DeepEnsemblePosteriorApproximator,
)
from fortuna.prob_model.posterior.map.map_posterior import MAPState
from fortuna.prob_model.posterior.map.map_trainer import (
    JittedMAPTrainer,
    MAPTrainer,
    MultiDeviceMAPTrainer,
)
from fortuna.prob_model.posterior.posterior_multi_state_repository import (
    PosteriorMultiStateRepository,
)
from fortuna.prob_model.posterior.run_preliminary_map import run_preliminary_map
from fortuna.typing import (
    Path,
    Status,
)
from fortuna.utils.builtins import get_dynamic_scale_instance_from_model_dtype
from fortuna.utils.device import select_trainer_given_devices
from fortuna.utils.freeze import get_trainable_paths
from fortuna.utils.nested_dicts import (
    nested_get,
    nested_set,
)

logger = logging.getLogger(__name__)


class DeepEnsemblePosterior(Posterior):
    def __init__(
        self,
        joint: Joint,
        posterior_approximator: DeepEnsemblePosteriorApproximator,
    ):
        """
        Deep ensemble approximate posterior class.

        Parameters
        ----------
        joint: Joint
            Joint distribution.
        posterior_approximator: DeepEnsemble
            Deep ensemble posterior approximator.
        """
        super().__init__(joint=joint, posterior_approximator=posterior_approximator)

    def __str__(self):
        return DEEP_ENSEMBLE_NAME

    def fit(
        self,
        train_data_loader: DataLoader,
        val_data_loader: Optional[DataLoader] = None,
        fit_config: FitConfig = FitConfig(),
        map_fit_config: Optional[FitConfig] = None,
        **kwargs,
    ) -> List[Status]:
        super()._checks_on_fit_start(fit_config, map_fit_config)

        status = dict()

        map_state = None
        if map_fit_config is not None and fit_config.optimizer.freeze_fun is None:
            logging.warning(
                "It appears that you are trying to configure `map_fit_config`. "
                "However, a preliminary run with MAP is supported only if "
                "`fit_config.optimizer.freeze_fun` is given. "
                "Since the latter was not given, `map_fit_config` will be ignored."
            )
        elif not super()._is_state_available_somewhere(
            fit_config
        ) and super()._should_run_preliminary_map(fit_config, map_fit_config):
            map_state, status["map"] = run_preliminary_map(
                joint=self.joint,
                train_data_loader=train_data_loader,
                val_data_loader=val_data_loader,
                map_fit_config=map_fit_config,
                rng=self.rng,
                **kwargs,
            )

        trainer_cls = select_trainer_given_devices(
            devices=fit_config.processor.devices,
            base_trainer_cls=MAPTrainer,
            jitted_trainer_cls=JittedMAPTrainer,
            multi_device_trainer_cls=MultiDeviceMAPTrainer,
            disable_jit=fit_config.processor.disable_jit,
        )

        train_data_size = train_data_loader.size
        val_data_size = val_data_loader.size if val_data_loader is not None else None

        def _fit(i):
            if self._is_state_available_somewhere(fit_config):
                _state = self._restore_state_from_somewhere(
                    i=i,
                    fit_config=fit_config,
                    allowed_states=(MAPState,),
                )
            else:
                _state = self._init_map_state(map_state, train_data_loader, fit_config)

            _state = self._freeze_optimizer_in_state(_state, fit_config)

            save_checkpoint_dir_i = (
                pathlib.Path(fit_config.checkpointer.save_checkpoint_dir) / str(i)
                if fit_config.checkpointer.save_checkpoint_dir
                else None
            )
            trainer = trainer_cls(
                predict_fn=self.joint.likelihood.prob_output_layer.predict,
                save_checkpoint_dir=save_checkpoint_dir_i,
                save_every_n_steps=fit_config.checkpointer.save_every_n_steps,
                keep_top_n_checkpoints=fit_config.checkpointer.keep_top_n_checkpoints,
                disable_training_metrics_computation=fit_config.monitor.disable_training_metrics_computation,
                eval_every_n_epochs=fit_config.monitor.eval_every_n_epochs,
                early_stopping_monitor=fit_config.monitor.early_stopping_monitor,
                early_stopping_min_delta=fit_config.monitor.early_stopping_min_delta,
                early_stopping_patience=fit_config.monitor.early_stopping_patience,
                freeze_fun=fit_config.optimizer.freeze_fun,
            )

            return trainer.train(
                rng=self.rng.get(),
                state=_state,
                loss_fun=self.joint._batched_negative_log_joint_prob,
                training_dataloader=train_data_loader,
                training_dataset_size=train_data_size,
                n_epochs=fit_config.optimizer.n_epochs,
                metrics=fit_config.monitor.metrics,
                validation_dataloader=val_data_loader,
                validation_dataset_size=val_data_size,
                verbose=fit_config.monitor.verbose,
                callbacks=fit_config.callbacks,
                max_grad_norm=fit_config.hyperparameters.max_grad_norm,
                gradient_accumulation_steps=fit_config.hyperparameters.gradient_accumulation_steps,
            )

        if isinstance(self.state, PosteriorMultiStateRepository):
            for i in range(self.posterior_approximator.ensemble_size):
                self.state.state[i].checkpoint_dir = (
                    pathlib.Path(fit_config.checkpointer.save_checkpoint_dir) / str(i)
                    if fit_config.checkpointer.save_checkpoint_dir is not None
                    and fit_config.checkpointer.dump_state
                    else None
                )
        else:
            self.state = PosteriorMultiStateRepository(
                size=self.posterior_approximator.ensemble_size,
                checkpoint_dir=fit_config.checkpointer.save_checkpoint_dir
                if fit_config.checkpointer.dump_state is True
                else None,
            )

        status = []
        for i in range(self.posterior_approximator.ensemble_size):
            logging.info(
                f"Run {i+1} out of {self.posterior_approximator.ensemble_size}."
            )
            state, _status = _fit(i)
            self.state.put(
                state=state, i=i, keep=fit_config.checkpointer.keep_top_n_checkpoints
            )
            status.append(_status)
        logging.info("Fit completed.")
        return status

    def sample(self, rng: Optional[PRNGKeyArray] = None, **kwargs) -> JointState:
        if rng is None:
            rng = self.rng.get()
        state = pure_callback(
            lambda j: self.state.get(i=j),
            self.state.get(i=0),
            random.choice(rng, self.posterior_approximator.ensemble_size),
        )
        return JointState(
            params=state.params,
            mutable=state.mutable,
            calib_params=state.calib_params,
            calib_mutable=state.calib_mutable,
        )

    def load_state(self, checkpoint_dir: Path) -> None:
        try:
            self.restore_checkpoint(pathlib.Path(checkpoint_dir) / "0")
        except ValueError:
            raise ValueError(
                f"No checkpoint was found in `checkpoint_dir={checkpoint_dir}`."
            )
        self.state = PosteriorMultiStateRepository(
            size=self.posterior_approximator.ensemble_size,
            checkpoint_dir=checkpoint_dir,
        )

    def save_state(self, checkpoint_dir: Path, keep_top_n_checkpoints: int = 1) -> None:
        for i in range(self.posterior_approximator.ensemble_size):
            self.state.put(state=self.state.get(i), i=i, keep=keep_top_n_checkpoints)

    def _init_map_state(
        self, state: Optional[MAPState], data_loader: DataLoader, fit_config: FitConfig
    ) -> MAPState:
        if state is None or fit_config.optimizer.freeze_fun is None:
            state = super()._init_joint_state(data_loader)

            return MAPState.init(
                params=state.params,
                mutable=state.mutable,
                optimizer=fit_config.optimizer.method,
                calib_params=state.calib_params,
                calib_mutable=state.calib_mutable,
                dynamic_scale=get_dynamic_scale_instance_from_model_dtype(
                    getattr(self.joint.likelihood.model_manager.model, "dtype")
                    if hasattr(self.joint.likelihood.model_manager.model, "dtype")
                    else None
                ),
            )
        else:
            random_state = super()._init_joint_state(data_loader)
            trainable_paths = get_trainable_paths(
                state.params, fit_config.optimizer.freeze_fun
            )
            state = state.replace(
                params=FrozenDict(
                    nested_set(
                        d=state.params.unfreeze(),
                        key_paths=trainable_paths,
                        objs=tuple(
                            [
                                nested_get(d=random_state.params, keys=path)
                                for path in trainable_paths
                            ]
                        ),
                    )
                )
            )

        return state

    def _restore_state_from_somewhere(
        self,
        i: int,
        fit_config: FitConfig,
        allowed_states: Optional[Tuple[Type[MAPState], ...]] = None,
    ) -> MAPState:
        if fit_config.checkpointer.restore_checkpoint_path is not None:
            restore_checkpoint_path = pathlib.Path(
                fit_config.checkpointer.restore_checkpoint_path
            ) / str(i)
            state = self.restore_checkpoint(
                restore_checkpoint_path=restore_checkpoint_path,
                optimizer=fit_config.optimizer.method,
            )
        elif fit_config.checkpointer.start_from_current_state is not None:
            state = self.state.get(i=i, optimizer=fit_config.optimizer.method)

        if allowed_states is not None and not isinstance(state, allowed_states):
            raise ValueError(
                f"The type of the restored checkpoint must be within {allowed_states}. "
                f"However, {fit_config.checkpointer.restore_checkpoint_path} pointed to a state "
                f"with type {type(state)}."
            )

        self._check_state(state)
        return state



from __future__ import annotations

from typing import (
    Any,
    Callable,
    Dict,
    Optional,
    Tuple,
    Union,
)

from flax.core import FrozenDict
from jax._src.prng import PRNGKeyArray
import jax.numpy as jnp
from optax._src.base import PyTree

from fortuna.prob_model.posterior.map import *
from fortuna.prob_model.posterior.map.map_state import MAPState
from fortuna.prob_model.posterior.posterior_trainer import PosteriorTrainerABC
from fortuna.training.trainer import (
    JittedMixin,
    MultiDeviceMixin,
)
from fortuna.typing import (
    Array,
    Batch,
    CalibMutable,
    CalibParams,
    Mutable,
    Params,
)


class MAPTrainer(PosteriorTrainerABC):
    def training_loss_step(
        self,
        loss_fun: Callable[[Any], Union[float, Tuple[float, dict]]],
        params: Params,
        batch: Batch,
        mutable: Mutable,
        rng: PRNGKeyArray,
        n_data: int,
        unravel: Optional[Callable[[any], PyTree]] = None,
        calib_params: Optional[CalibParams] = None,
        calib_mutable: Optional[CalibMutable] = None,
        kwargs: FrozenDict[str, Any] = FrozenDict(),
    ) -> Tuple[jnp.ndarray, Dict[str, Any]]:
        return_aux = ["outputs"]
        if mutable is not None:
            return_aux += ["mutable"]
        loss, aux = loss_fun(
            params,
            batch,
            n_data=n_data,
            mutable=mutable,
            return_aux=return_aux,
            train=True,
            rng=rng,
            calib_params=calib_params,
            calib_mutable=calib_mutable,
        )
        logging_kwargs = None
        return (
            loss,
            {
                "outputs": aux.get("outputs"),
                "mutable": aux.get("mutable"),
                "logging_kwargs": logging_kwargs,
            },
        )

    def validation_step(
        self,
        state: MAPState,
        batch: Batch,
        loss_fun: Callable[[Any], Union[float, Tuple[float, dict]]],
        rng: PRNGKeyArray,
        n_data: int,
        metrics: Optional[Tuple[Callable[[jnp.ndarray, Array], float], ...]] = None,
        unravel: Optional[Callable[[any], PyTree]] = None,
        kwargs: FrozenDict[str, Any] = FrozenDict(),
    ) -> Dict[str, jnp.ndarray]:
        loss, aux = loss_fun(
            self._get_all_params(state),
            batch,
            n_data=n_data,
            mutable=state.mutable,
            return_aux=["outputs"],
            train=False,
            calib_params=state.calib_params,
            calib_mutable=state.calib_mutable,
        )

        if metrics is not None:
            val_metrics = self.compute_metrics(
                self.predict_fn(aux["outputs"]), batch[1], metrics
            )
            return {
                "val_loss": loss,
                **{f"val_{m}": v for m, v in val_metrics.items()},
            }
        return dict(val_loss=loss)

    def __str__(self):
        return MAP_NAME


class JittedMAPTrainer(JittedMixin, MAPTrainer):
    pass


class MultiDeviceMAPTrainer(MultiDeviceMixin, MAPTrainer):
    pass



from typing import (
    Any,
    Callable,
    Tuple,
)

import jax
import jax.numpy as jnp
import jax.scipy as jsp

from fortuna.typing import Array


def cross_entropy_loss_fn(
    outputs: jnp.ndarray,
    targets: Array,
) -> jnp.ndarray:
    """
    A cross-entropy loss function. Check `here <https://en.wikipedia.org/wiki/Cross_entropy>`_ for reference.

    Parameters
    ----------
    outputs: Array
        Model outputs to be passed to the loss.
    targets: Array
        Target data points.

    Returns
    -------
    Tuple[jnp.ndarray, Any]
        The cross-entropy loss evaluation and auxiliary objects.
    """
    targets = jax.nn.one_hot(targets, outputs.shape[-1])
    return jnp.mean(jsp.special.logsumexp(outputs, -1) - jnp.sum(targets * outputs, -1))



import flax.linen as nn
import jax.numpy as jnp

from fortuna.typing import Array


class HyperparameterModel(nn.Module):
    r"""
    A hyperparameter model. The value of the hyperparameter will not change during training.

    Parameters
    ----------
    value: Union[float, Array]
        Value of the hyperparameter.
    """
    value: Array

    def setup(self) -> None:
        if self.value.ndim != 1:
            raise ValueError(
                "`value` must be a one-dimensional array, with length equal to the output dimension of "
                "the model."
            )
        dummy = self.param("none", nn.initializers.zeros, (0,))

    def __call__(self, x: Array, **kwargs) -> jnp.ndarray:
        return jnp.broadcast_to(self.value, shape=(x.shape[0], len(self.value)))
